"""Console script for timelink.

Also provides basic mhk manager functionality.

Run with  python -m timelink.cli

"""

import platform
import uvicorn
import typer
import docker
from timelink.mhk.utilities import get_mhk_info, is_mhk_installed

server: uvicorn.Server = None  # uvicorn server instance

# We use Typer https://typer.tiangolo.com
app = typer.Typer(help="Timelink and MHK manager")
mhk_app = typer.Typer()
app.add_typer(mhk_app, name="mhk", help="MHK legacy manager")


# Start server
# see alternative methods
# https://github.com/miguelgrinberg/python-socketio/issues/332#issuecomment-712928157
# https://stackoverflow.com/questions/68603658/how-to-terminate-a-uvicorn-fastapi-application-cleanly-with-workers-2-when


@app.command("start")
def start():
    """Starts timelink with uvicorn"""
    typer.echo("Starting Timelink")
    config = uvicorn.Config("timelink.app.main:app", port=8008, reload=True)

    global server
    server = uvicorn.Server(config)
    server.run()
    return 0


@mhk_app.command(name="version")
def mhk_version():
    """shows MHK manager version

    Demonstrates how to access MHK installation files and
    usage of Docker API
    """
    if is_mhk_installed():
        mhk_info = get_mhk_info()

    try:
        client = docker.from_env(version="auto")
        dv = client.version()
        mhkv = f"""    Manager version:  {mhk_info.mhk_version}
        Docker version:   {dv["Version"]}
        Host OS:          {platform.system()} {platform.release()}
        User home:        {mhk_info.user_home}
        mhk-home:         {mhk_info.mhk_home}
        mhk-home init:    {mhk_info.mhk_home_init}
        mhk-home update:  {mhk_info.mhk_home_update}
        mhk use-tag:      {mhk_info.mhk_app_env.get("TAG", "*none*")}
        mhk local host:   {mhk_info.mhk_host}
        MHK URL:          http://127.0.0.1:8080/mhk
        Kleio URL:        http://127.0.0.1:8088
        Portainer URL:    http://127.0.0.1:9000"""
        typer.echo(mhkv)
    except Exception as e:
        typer.echo(f"Could not access docker: {e}")
    else:
        type.echo("Could not find a MHK instalation")
    return 0


@mhk_app.command(name="status")
def mhk_status():
    """shows docker status information"""
    client = docker.from_env()
    dinfo = client.info()
    typer.echo(
        f"""
    Containers  :{dinfo['Containers']}
       Running  :{dinfo['ContainersRunning']}
       Paused   :{dinfo['ContainersPaused']}
       Stopped  :{dinfo['ContainersStopped']}
    """
    )
    return 0


@app.callback()
def main():
    """
    This is the timelink/MHK manager on the command line
    """

    typer.echo("This is the timelink/MHK manager")


if __name__ == "__main__":
    app()  # pragma: no cover



""" Handling views from sql Alchemy

    View utilities from https://github.com/sqlalchemy/sqlalchemy/wiki/Views
"""

import sqlalchemy as sa
from sqlalchemy.ext import compiler
from sqlalchemy.schema import DDLElement
from sqlalchemy.sql import table


class CreateView(DDLElement):
    """Create a View"""

    def __init__(self, name, selectable):
        self.name = name
        self.selectable = selectable


class DropView(DDLElement):
    """Drop a View"""

    def __init__(self, name):
        self.name = name


@compiler.compiles(CreateView)
def _create_view(element, compiler, **kw):
    return "CREATE VIEW %s AS %s" % (
        element.name,
        compiler.sql_compiler.process(element.selectable, literal_binds=True),
    )


@compiler.compiles(DropView)
def _drop_view(element, compiler, **kw):
    return "DROP VIEW %s" % (element.name)


def view_exists(ddl, target, connection, **kw):
    return ddl.name in sa.inspect(connection).get_view_names()


def view_doesnt_exist(ddl, target, connection, **kw):
    return not view_exists(ddl, target, connection, **kw)


def view(name, metadata, selectable):
    """Create a view with the given name from the given selectable.

    The view is created when the metadata is first bound to an engine.

    Example:

    .. code-block:: python

       stuff_view
            = view("stuff_view", metadata, sa.select(
                stuff.c.id.label("id"),
                stuff.c.data.label("data"),
                more_stuff.c.data.label("moredata"),
            )
                .select_from(stuff.join(more_stuff))
                .where(stuff.c.data.like(("%orange%"))),
            )

        with engine.connect() as conn:
        conn.execute(
            sa.select(stuff_view.c.data, stuff_view.c.moredata)
        ).all()

    """

    t = table(name)

    t._columns._populate_separate_keys(
        col._make_proxy(t) for col in selectable.selected_columns
    )

    sa.event.listen(
        metadata,
        "after_create",
        CreateView(name, selectable).execute_if(callable_=view_doesnt_exist),
    )
    sa.event.listen(
        metadata, "before_drop", DropView(name).execute_if(callable_=view_exists)
    )
    return t



"""Unit test package for timelink."""
# pylint: disable=unused-import, import-error
import os
from pathlib import Path
import random
import time
import warnings

from sqlalchemy.orm import sessionmaker
import pytest

from timelink.kleio.kleio_server import KleioServer
from timelink.kleio.schemas import KleioFile

# See https://docs.sqlalchemy.org/en/14
#      /orm/session_basics.html#when-do-i-make-a-sessionmaker
# Tests should import from here to access the session
#

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
# conn_string local file version
# from pathlib import Path
# sqlite_db = Path(TEST_DIR, "/sqlite/test.db")
# conn_string = f'sqlite://{sqlite_db}?check_same_thread=False'

# con_string sqlite in memory version
conn_string = "sqlite://?check_same_thread=False"

Session = sessionmaker()

mhk_absent = pytest.mark.skipif(
    not Path(Path.home(), ".mhk").is_file(), reason="MHK not present"
)

skip_on_travis = pytest.mark.skipif(
    os.environ.get('TRAVIS') == 'true',
    reason="this test requires file system access for sqlite"
)


def get_one_translation(kserver: KleioServer,
                        path="",
                        max_wait=120) -> KleioFile:
    """Get one translation from the server"""
    translations = kserver.translation_status(path=path, recurse="yes", status="V")
    if len(translations) == 0:
        need_translation = kserver.translation_status(path=path, recurse="yes", status="T")
        if len(need_translation) == 0:
            raise RuntimeError(f"No files available for translation, path={path}")

        one_translation = random.choice(need_translation)
        kserver.translate(path=one_translation.path, recurse="no")

        counter = 0
        wait_for = 1
        max_counter = max_wait / wait_for
        translations = kserver.translation_status(path=path, recurse="yes", status="V")
        in_process = kserver.translation_status(path=path, recurse="yes", status="P")
        while len(translations) == 0 and (counter < max_counter
                                          or len(in_process) > 0):  # noqa: W503
            time.sleep(wait_for)
            counter += 1
            if counter % 20 == 0:
                warnings.warn(f"Waiting for translations {one_translation.name}, counter={counter}",
                              stacklevel=1)
            translations = kserver.translation_status(path=path, recurse="yes", status="V")
            in_process = kserver.translation_status(path=path, recurse="yes", status="P")

    if len(translations) > 0:
        kleio_file: KleioFile = random.choice(translations)
    else:
        raise RuntimeError("No translations available in server")

    return kleio_file



""".. module:: utilities
   :synopsis: Various utilities for handling Kleio groups and elements

.. moduleauthor: Joaquim Ramos de Carvalho

Kleio Groups are the building blocks for transcription of historical sources.
"""

import textwrap
from os import linesep as nl


def kleio_escape(v: str) -> str:
    """
    Checks for Kleio special characters and quotes if needed::

         >>> print(kleio_escape('normal string'))
         normal string
         >>> print(kleio_escape('oops we have a / in the middle'))
         "oops we have a / in the middle"

    """
    if v is None:
        return None
    s = str(v)
    if any(i in s for i in "/;=$#%\n"):
        return '"' + s + '"'
    else:
        return s


def quote_long_text(txt, initial_indent=" " * 4, indent=" " * 2, width=2048) -> str:
    """Surround long text with triple quotes,
    wraps and indents lines if needed.
    Some of the parameters are passed on to :py:func:`textwrap.fill`.

    Sphynx style markup

    :param txt: The text to be transformed
    :type txt: str
    :param initial_indent: string to ident the first line of paragraphs.
                        Default is 4 spaces. See :py:func:`textwrap.fill`.
    :type initial_indent: str
    :param indent: string to ident the wrap lines of
                    paragraphs (after the first).
                    Default is 2 spaces. See :py:func:`textwrap.fill`.
    :type indent: str
    :param width: width of line for wrapping. See :py:func:`textwrap.fill`.
    :type width: int
    :rtype: str
    """
    if txt is None:
        return None
    if len(txt) > 127 or len(txt.splitlines()) > 1:
        s = '"""' + nl
        for line in txt.splitlines():
            w = textwrap.fill(line, width=width, initial_indent=initial_indent)
            s = s + textwrap.indent(w, indent) + nl
        s = s + indent + '"""'
    else:
        s = kleio_escape(txt)
    return s



"""Test the import of Kleio files into the Timelink database.
Check tests.__init__.py for parameters

"""
# pylint: disable=import-error
import logging
import os
from pathlib import Path
from time import sleep

import pytest

from tests import TEST_DIR, get_one_translation, skip_on_travis
from timelink.api.models.system import KleioImportedFile
from timelink.kleio.importer import import_from_xml
from timelink.api.models import base  # pylint: disable=unused-import. # noqa: F401
from timelink.api.models.base import Person
from timelink.api.database import (
    TimelinkDatabase,
    start_postgres_server,
    get_postgres_container_user,
    get_postgres_container_pwd,
    get_import_status,
)
from timelink.api.models.entity import Entity

from timelink.kleio import KleioServer
from timelink.kleio.schemas import KleioFile

# https://docs.pytest.org/en/latest/how-to/skipping.html
pytestmark = skip_on_travis


@pytest.fixture(scope="module")
def dbsystem(request):
    """Create a database for testing"""
    db_type, db_name, db_url, db_user, db_pwd = request.param
    if db_type == "postgres":
        start_postgres_server()
        db_user = get_postgres_container_user()
        db_pwd = get_postgres_container_pwd()

    database = TimelinkDatabase(db_name, db_type, db_url, db_user, db_pwd)
    try:
        yield database
    finally:
        # database.drop_db
        database.session().close()


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        # change to pytest.param("postgres", "tests", None, None, None, marks=skip_on_travis)
        # to skip the test on travis see https://doc.pytest.org/en/latest/how-to/skipping.html#skip-xfail-with-parametrize
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_xml(dbsystem):
    """Test the import of a Kleio file into the Timelink database"""
    file: Path = Path(TEST_DIR, "xml_data/b1685.xml")
    session = dbsystem.session()
    try:
        stats = import_from_xml(file, session, options={"return_stats": True})
    except Exception as exc:
        print(exc)
        raise
    sfile: Path = stats["file"]
    assert sfile.name == file.name
    domingos_vargas = session.get(Person, "b1685.33-per6")
    assert domingos_vargas is not None, "could not get a person from file"
    kleio = domingos_vargas.to_kleio()
    assert len(kleio) > 0


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_linked_data_attributes(dbsystem):
    """Test the import of a translation with linked_data in attributes"""
    file: Path = Path(TEST_DIR, "xml_data/dehergne-a.xml")
    session = dbsystem.session()
    try:
        stats = import_from_xml(file, session, options={"return_stats": True})
    except Exception as exc:
        print(exc)
        raise
    sfile: Path = stats["file"]
    assert sfile.name == file.name
    per: Person = session.get(Person, "deh-antonio-de-abreu")
    assert per is not None, "could not get a group with linked data from file"
    # check for attributes with extra_info
    for a in per.attributes:
        if 'extra_info' in a.obs:
            assert a.get_extra_info() is not None


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_linked_data_geoentites(dbsystem):
    """Test the import of a translation with linked_data"""
    file: Path = Path(TEST_DIR, "xml_data/dehergne-locations-1644.xml")
    session = dbsystem.session()
    try:
        stats = import_from_xml(file, session, options={"return_stats": True})
    except Exception as exc:
        print(exc)
        raise
    sfile: Path = stats["file"]
    assert sfile.name == file.name
    geo1 = session.get(Entity, "deh-r1644-chekiang")
    assert geo1 is not None, "could not get a group with linked data from file"
    assert geo1.get_extra_info() is not None


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_atr_date(dbsystem):
    """Test the import of a Kleio file with explicit attribute dates"""
    file: Path = Path(TEST_DIR, "xml_data/test-atr-date.xml")
    session = dbsystem.session()
    try:
        stats = import_from_xml(file, session, options={"return_stats": True})
    except Exception as exc:
        print(exc)
        raise
    sfile: Path = stats["file"]
    assert sfile.name == file.name
    domingos_vargas = session.get(Person, "t140337")
    assert domingos_vargas is not None, "could not get a person from file"
    kleio = domingos_vargas.to_kleio()
    assert len(kleio) > 0


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_with_custom_mapping(dbsystem):
    """Test the import of a Kleio file into the Timelink database with a custom mapping"""
    file = Path(TEST_DIR, "xml_data/dev1692.xml")
    with dbsystem.session() as session:
        try:
            stats = import_from_xml(file, session, options={"return_stats": True})
        except Exception as exc:
            print(exc)
            raise
        sfile = stats["file"]
        assert "dev1692" in sfile.name
        caso = session.get(Entity, "c1692-antonio-cordeiro-alcouc")
        caso_kleio = caso.to_kleio()
        assert len(caso_kleio) > 0
        assert caso is not None, "could not get an entity from special mapping"
        caso_kleio = caso.to_kleio()
        assert len(caso_kleio) > 0
        test = session.get(Entity, "dev1692-per5")
        assert test is not None
        test_kleio = test.to_kleio()
        assert len(test_kleio) > 0


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_with_many(dbsystem):
    """Test the import of a long Kleio file into the Timelink database"""
    file = Path(TEST_DIR, "xml_data/test-auc-alunos-264605-A-140337-140771.xml")
    with dbsystem.session() as session:
        try:
            stats = import_from_xml(file, session, options={"return_stats": True})
        except Exception as exc:
            print(exc)
            raise
        sfile = stats["file"]
        assert "alunos" in sfile.name
        estudante = session.get(Entity, "140771")
        kleio = estudante.to_kleio()
        assert len(kleio) > 0
        assert estudante is not None, (
            "could not get an entity from big import"
        )  # noqa


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_identifications(dbsystem):
    """Test the import a identifications file"""
    file = Path(TEST_DIR, "xml_data/mhk_identification_toliveira.xml")
    with dbsystem.session() as session:
        try:
            stats = import_from_xml(file, session, options={"return_stats": True})
        except Exception as exc:
            print(exc)
            raise
        sfile = stats["file"]
        assert "identification" in sfile.name
        real_person = session.get(Entity, "rp-46")
        kleio = real_person.to_kleio()
        assert len(kleio) > 0
        assert real_person is not None, (
            "could not get a real person from identifications import"
        )  # noqa


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_git_hub(dbsystem):
    """Test the import of a Kleio file from github into the Timelink database"""
    file = "https://raw.githubusercontent.com/time-link/timelink-py/f76007cb7b98b39b22be8b70b3b2a62e7ae0c12f/tests/xml_data/b1685.xml"  # noqa
    with dbsystem.session() as session:
        try:
            stats = import_from_xml(file, session, options={"return_stats": True})
        except Exception as exc:
            print(exc)
            raise
        domingos_vargas = session.get(Person, "b1685.33-per6")
        assert domingos_vargas is not None, "could not get a person from file"
        kleio = domingos_vargas.to_kleio()
        assert len(kleio) > 0
        sfile: Path = stats["file"]
        assert "b1685.xml" in sfile
        pass


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_from_kleio_server(dbsystem):
    """Test the import of a Kleio file from kleio server into the Timelink database"""
    kserver: KleioServer = KleioServer.start(kleio_home=f"{TEST_DIR}/timelink-home",
                                             reuse=True,
                                             update=False)
    kleio_url = kserver.get_url()
    kleio_token = kserver.get_token()
    # wait for the server to start
    server_up = False
    counter = 0
    while not server_up and counter < 10:
        try:
            kserver.get_home_page()
            server_up = True
        except Exception as exc:
            logging.debug(exc)
            sleep(1)
            counter += 1
            pass
    kleio_file: KleioFile = get_one_translation(kserver)
    file = kleio_file.xml_url
    with dbsystem.session() as session:
        try:
            import_from_xml(
                file,
                session,
                options={
                    "return_stats": True,
                    "kleio_token": kleio_token,
                    "kleio_url": kleio_url,
                    "mode": "TL",
                },
            )
        except Exception as exc:
            print(exc)

        # check if the file was imported
        filename_with_extension = os.path.basename(file)
        filename, extension = os.path.splitext(filename_with_extension)
        sources = session.query(KleioImportedFile).all()
        imported = False
        for source in sources:
            kfilename_with_extension = os.path.basename(source.path)
            kfile, kextention = os.path.splitext(kfilename_with_extension)
            if kfile == filename:
                imported = True
                break

        assert imported, "file not imported"


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_status_1(dbsystem):
    """Test if import status is retrieved"""
    kserver: KleioServer = KleioServer.start(kleio_home=f"{TEST_DIR}/timelink-home",
                                             reuse=True,
                                             update=False)
    kleio_file = get_one_translation(kserver)
    translations = [kleio_file]
    assert len(translations) > 0, "no valid translations found in Kleio Server"
    import_status = get_import_status(dbsystem, translations, match_path=True)
    assert import_status is not None


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_import_sources_with_no_year(dbsystem):
    """Test the import of a Kleio file from kleio server into the Timelink database"""
    kserver: KleioServer = KleioServer.start(kleio_home=f"{TEST_DIR}/timelink-home")
    kleio_url = kserver.get_url()
    kleio_token = kserver.get_token()
    kleio_file: KleioFile = get_one_translation(kserver, path="")
    file = kleio_file.xml_url
    # file: KleioFile
    # temp = [
    #   kfile.xml_url
    #    for kfile in translations
    #   if kfile.name == "auc-alunos.cli"
    # ]
    # file = temp[0]
    with dbsystem.session() as session:
        try:
            import_from_xml(
                file,
                session,
                options={
                    "return_stats": True,
                    "kleio_token": kleio_token,
                    "kleio_url": kleio_url,
                    "mode": "TL",
                },
            )
        except Exception as exc:
            print(exc)

        # check if the file was imported
        filename_with_extension = os.path.basename(file)
        filename, extension = os.path.splitext(filename_with_extension)
        sources = session.query(KleioImportedFile).filter_by().all()
        imported = False
        for source in sources:
            kfilename_with_extension = os.path.basename(source.path)
            kfile, kextention = os.path.splitext(kfilename_with_extension)
            if kfile == filename:
                imported = True
                break

        assert imported, "file not imported"



""" Implementation of Acts in POM model

(c) Joaquim Carvalho 2023.
MIT License, no warranties.
"""

# for sqlalchemy 2.0 ORM
# see https://docs.sqlalchemy.org/en/20/orm/declarative_config.html
# and https://docs.sqlalchemy.org/en/20/orm/declarative_tables.html#orm-declarative-table

from sqlalchemy import ForeignKey  # pylint: disable=import-error
from sqlalchemy import String  # pylint: disable=import-error
from sqlalchemy.orm import Mapped  # pylint: disable=import-error
from sqlalchemy.orm import mapped_column  # pylint: disable=import-error

from timelink.kleio.utilities import quote_long_text, kleio_escape

from .entity import Entity


class Act(Entity):
    """Represents an Act, i.e. a record of some event in a historical document.

    Examples of acts are: baptisms, marriages, purchase deeds, wills,
    court records.

    TODO: implement https://github.com/time-link/timelink-kleio/issues/1
    """

    __tablename__ = "acts"

    #: str:  a string uniquely identifying the act
    id: Mapped[str] = mapped_column(
        String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True
    )
    the_type: Mapped[str] = mapped_column(
        String(32), nullable=True, comment="type of act"
    )
    #: str: the date of the act in Kleio format AAAAMMDD
    the_date: Mapped[str] = mapped_column(
        String,
        index=True,
        nullable=True,
        comment="the date of the act in Kleio format AAAAMMDD",
    )
    #: str: location of the act, eg church, notary office
    loc: Mapped[str] = mapped_column(String, nullable=True)
    ref: Mapped[str] = mapped_column(String, nullable=True)  #: str: archival reference
    obs: Mapped[str] = mapped_column(
        String, nullable=True
    )  #: any observations or comments.

    __mapper_args__ = {"polymorphic_identity": "act"}

    def __repr__(self):
        """Expression that can be used to instantiate an Act"""
        sr = super().__repr__()
        return (
            f"Act(id={sr}, "
            f'the_type="{kleio_escape(self.the_type)}", '
            f'the_date="{self.the_date}", '
            f'local="{kleio_escape(self.loc)}", '
            f'ref="{kleio_escape(self.ref)}", '
            f'obs="{quote_long_text(self.obs)}"'
            f")"
        )

    def __str__(self):
        r = (
            f"{self.groupname}${self.id}"
            f"/{self.the_date}"
            f"/type={kleio_escape(self.the_type)}"
            f"/ref={quote_long_text(self.ref)}"
            f"/loc={quote_long_text(self.loc)}"
        )
        if self.obs is not None:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r



"""Timelink Python package.

Timelink, formerly known as MHK (Micro History with Kleio)
is an information system
designed for processing person related
information collected from historical sources.

.. moduleauthor:: Joaquim Ramos de Carvalho



"""
from .api import database  # noqa: F401
from .api import models  # noqa: F401
from .api import views  # noqa: F401
from .api import schemas  # noqa: F401


__author__ = """Joaquim Ramos de Carvalho"""
__email__ = 'joaquim@uc.pt'
__version__ = '1.1.5'

version = __version__



from datetime import datetime

from sqlalchemy import (
    Column,
    String,
    Integer,
    DateTime,
    ForeignKey,
)  # pylint: disable=import-error
from sqlalchemy.orm import relationship, backref  # pylint: disable=import-error

from timelink.kleio.utilities import kleio_escape
from timelink.mhk.models.base_class import Base


class Entity(Base):
    """ORM Model root of the object hierarchy.

     All entities in a Timelink/MHK database have an entry in this table.
     Each entity is associated with a class that allow access to a
     specialization table with more columns for that class.

    This corresponds to the model described as "Joined Table Inheritance"
    in sqlalchemy (see https://docs.sqlalchemy.org/en/14/orm/inheritance.html)
    """

    __tablename__ = "entities"

    #: str: unique identifier for the entity
    id = Column(String, primary_key=True)
    #: str: name of the class. Links to pom_som_mapper class
    pom_class = Column(
        "class", String, ForeignKey("classes.id", use_alter=True), index=True
    )
    #: str: id of the entity inside which this occurred.
    inside = Column(String, ForeignKey("entities.id", ondelete="CASCADE"), index=True)
    #: int: sequential order of this entity in the source
    the_order = Column(Integer)
    #: int: the nesting level of this entity in the source
    the_level = Column(Integer)
    #: int: line in which the entity occurred in the source
    the_line = Column(Integer)
    #: str: name of the kleio group that produced this entity
    groupname = Column(String, index=True)
    #: datetime: when this entity was updated in the database
    updated = Column(DateTime, default=datetime.utcnow, index=True)
    #: datetime: when this entity was added to the full text index
    indexed = Column(DateTime, index=True)

    # These are defined in relation.py
    # rels_in = relationship("Relation", back_populates="dest")
    # rels_out = relationship("Relation", back_populates="org")
    # this based on
    # https://stackoverflow.com/questions/28843254
    #: list(Entity): list of Entity objects contained in this entity
    contains = relationship(
        "Entity",
        backref=backref("contained_by", remote_side="Entity.id"),
        cascade="all",
    )

    # see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    # To handle non mapped pom_class
    #      see https://github.com/sqlalchemy/sqlalchemy/issues/5445
    #
    #    __mapper_args__ = {
    #       "polymorphic_identity": "entity",
    #    "polymorphic_on": case(
    #        [(type.in_(["parent", "child"]), type)], else_="entity"
    #    ),
    #
    #  This defines what mappings do exist
    # [aclass.__mapper_args__['polymorphic_identity']
    #              for aclass in Entity.__subclasses__()]

    __mapper_args__ = {
        "polymorphic_identity": "entity",
        "polymorphic_on": pom_class,
    }

    @classmethod
    def get_subclasses(cls):
        """Get the subclasses of Entity"""
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def get_orm_entities_classes(cls):
        """Currently defined ORM classes that extend Entity
         (including Entity itself)


        Returns:
             list: List of ORM classes
        """
        sc = list(Entity.get_subclasses())
        sc.append(Entity)
        return sc

    @classmethod
    def get_som_mapper_ids(cls):
        """Ids of SomPomMapper references by orm classes

        Returns:
             List[str]: List of strings
        """
        return [
            aclass.__mapper_args__["polymorphic_identity"]
            for aclass in Entity.get_orm_entities_classes()
        ]

    @classmethod
    def get_tables_to_orm_as_dict(cls):
        """
        Return a dict with table name as key and ORM class as value
        """
        return {
            ormclass.__mapper__.local_table.name: ormclass
            for ormclass in Entity.get_orm_entities_classes()
        }

    @classmethod
    def get_som_mapper_to_orm_as_dict(cls):
        """
        Return a dict with pom_class id as key and ORM class as value
        """
        sc = Entity.get_orm_entities_classes()
        return {ormclass.__mapper__.polymorphic_identity: ormclass for ormclass in sc}

    @classmethod
    def get_orm_for_table(cls, table: String):
        """
        Entity.get_orm_for_table("acts")

        will return the ORM class handling the "acts" table
        """
        return cls.get_tables_to_orm_as_dict().get(table, None)

    @classmethod
    def get_orm_for_pom_class(cls, pom_class: str):
        """
        Entity.get_orm_for_pom_class("act")

        will return the ORM class corresponding to the pom_class "act"
        """
        return cls.get_som_mapper_to_orm_as_dict().get(pom_class, None)

    @classmethod
    def get_entity(cls, id: str, session=None):
        """
        Get an Entity from the database. The object returned
        will be of the ORM class defined by mappings.
        :param id: id of the entity
        :param session: current session
        :return: an Entity object of the proper class for the mapping
        """
        entity = session.get(Entity, id)
        if entity is not None:
            if entity.pom_class != "entity":
                orm_class = Entity.get_orm_for_pom_class(entity.pom_class)
                object_for_id = session.get(orm_class, id)
                return object_for_id
            else:
                return entity
        else:
            return None

        return entity

    def __repr__(self):
        return (
            f'Entity(id="{self.id}", '
            f'pom_class="{self.pom_class}",'
            f'inside="{self.inside}", '
            f"the_order={self.the_order}, "
            f"the_level={self.the_level}, "
            f"the_line={self.the_line}, "
            f'groupname="{self.groupname}", '
            f"updated={self.updated}, "
            f"indexed={self.indexed},"
            f")"
        )

    def __str__(self):
        return f"{self.groupname}${kleio_escape(self.id)}/type={kleio_escape(self.pom_class)}"

    def to_kleio(self, ident="", ident_inc="  "):
        s = f"{ident}{str(self)}"
        for inner in self.contains:
            innerk = inner.to_kleio(ident=ident + ident_inc, ident_inc=ident_inc)
            s = f"{s}\n{innerk}"
        return s



from typing import Optional
from datetime import datetime
from pydantic import BaseModel as BaseModel
from pydantic import ConfigDict
from timelink.kleio.kleio_server import KleioServer  # noqa
from timelink.api.database import TimelinkDatabase  # noqa
from ..schemas import UserSchema


class ProjectAccess(BaseModel):
    id: int
    user: UserSchema
    project: "ProjectSchema"
    access_level: Optional[str] = None

    # ORM mode
    model_config = ConfigDict(from_attributes=True)


class ProjectSchema(BaseModel):
    id: int
    name: str
    description: Optional[str] = None
    databaseURL: Optional[str] = None
    kleioServerURL: Optional[str] = None
    created: Optional[datetime] = None
    updated: Optional[datetime] = None

    # ORM mode
    model_config = ConfigDict(from_attributes=True)



# pylint: disable=import-error

from sqlalchemy import Column, String, ForeignKey

from timelink.mhk.models.entity import Entity
from timelink.kleio.utilities import quote_long_text


class Person(Entity):
    """Represents a person."""

    __tablename__ = "persons"

    id = Column(String, ForeignKey("entities.id"), primary_key=True)
    name = Column(String, index=True)
    id = Column(String, ForeignKey("entities.id"), primary_key=True)
    name = Column(String)
    sex = Column(String(1))
    obs = Column(String)

    __mapper_args__ = {"polymorphic_identity": "person"}

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Person(id={sr}, "
            f'name="{self.name}", '
            f'sex="{self.sex}", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        r = f"{self.groupname}${quote_long_text(self.name)}/{quote_long_text(self.sex)}/id={quote_long_text(self.id)}"
        if self.obs is not None:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r



import enum
from typing import List, Optional
from pydantic import BaseModel, Field, create_model
from fastapi import Request
from fastui import components as c


from timelink.app.backend.timelink_webapp import TimelinkWebApp
from timelink.app.schemas.user import UserSchema, UserProjectSchema
from .home_page import home_page


class ProjectInfoSchema(BaseModel):
    """Classe to hold the relevant project info for display."""

    id: int
    name: str
    description: Optional[str] = None
    databaseURL: Optional[str] = None
    kleioServerURL: Optional[str] = None


async def projects_info(
    webapp: TimelinkWebApp, request: Request, user: UserSchema
) -> c.Page:
    if user is not None:
        user_projects: List[UserProjectSchema] = user.projects
    else:
        user_projects = []

    ProjectEnumDynamic = enum.Enum(
        "ProjectEnum", [(p.project_name, p.project_name.lower()) for p in user_projects]
    )

    SelectFormDynamic = create_model(
        "SelectFormDynamic",
        project_name=(ProjectEnumDynamic, Field(title="Select project")),
    )
    # selectList: List[str] = enum.Enum('selectEnum', [name for name in user_projects])
    post_url = request.url_for("project_select")
    current_project = user.current_project_name
    if current_project is None:
        current_project = "None"
    if user is None:
        user_name = "guest"
    else:
        user_name = user.name
        if user_name == "":
            user_name = user.email.split("@")[0]
    return await home_page(
        c.Page(
            components=[
                c.Heading(level=6, text=f"{user_name}; current project: {current_project}"),
                c.Table(
                    data=user_projects,
                    data_model=UserProjectSchema,
                    no_data_message="No projects found. Associate users with projects in the admin page.",
                ),
                c.Div(
                    components=[
                        c.ModelForm(
                            model=SelectFormDynamic,
                            display_mode="page",
                            submit_url=str(post_url),
                            class_name="text-align: left"
                        )
                    ],
                    class_name='text-align: left'
                ),
            ]
        ),
        request=request,
        title="Project list",
        user=user,
    )



""" tests for interfacing to kleio server

https://github.com/time-link/timelink-kleio-server
"""

from enum import Enum
import time
import pytest

from tests import TEST_DIR
import timelink.kleio.kleio_server as kleio_server
from timelink.kleio.schemas import KleioFile, TokenInfo
from timelink.kleio.kleio_server import KleioServer, KleioServerForbidenException
import re

KLEIO_ADMIN_TOKEN: str = None
KLEIO_LIMITED_TOKEN: str = None
KLEIO_NORMAL_TOKEN: str = None

KLEIO_SERVER: KleioServer = None


class KleioServerTestMode(Enum):
    LOCAL = "local"
    DOCKER = "docker"


mode = KleioServerTestMode.DOCKER


@pytest.fixture(scope="function", autouse=True)
def setup():
    """setup kleio server for tests

    To run tests with a local Kleio Server outside docker:
    1) set "mode" abose to KleioServerTestMode.LOCAL
    2)Run the server in Prolog loading serverStart.pl and then:

        setenv('KLEIO_ADMIN_TOKEN','mytoken').
        setup_and_run_server(run_debug_server,[port(8089)]).

    """
    if mode == KleioServerTestMode.LOCAL:
        token = "mytoken"
        url = "http://localhost:8089"
        ks = KleioServer.attach(url, token)
    else:
        """Setup kleio server for tests"""
        khome = f"{TEST_DIR}/timelink-home"
        ks = KleioServer.start(kleio_home=khome, reuse=True, update=False)

    return ks


def test_find_kleio_home() -> str:
    """Test if kleio home is found"""
    kleio_home = KleioServer.find_local_kleio_home()
    assert kleio_home is not None


def test_is_kleio_server_running():
    """Test if kleio server is running"""
    ks = KleioServer.get_server()
    assert ks is not None or ks is None


def test_start_kleio_server():
    """Test if kleio server is started"""
    ks = KleioServer.start(kleio_home=f"{TEST_DIR}/timelink-home", update=False)
    assert ks is not None


def test_attach_kleio_server(setup):
    """Test attach to a running kleio server"""
    kserver: KleioServer = setup
    token = kserver.get_token()
    url = kserver.get_url()
    kleio_home = kserver.get_kleio_home()

    ks = KleioServer.attach(url, token, kleio_home)
    assert ks is not None


def test_get_kleio_server_container(setup):
    kserver: KleioServer = setup
    """Test get the container of the running kleio server"""
    container = kserver.get_container()
    assert container is not None


def test_get_kleio_server_token(setup):
    kserver: KleioServer = setup
    """Test if kleio server token is available"""
    assert kserver.get_token() is not None


def test_kleio_get_url(setup):
    kserver: KleioServer = setup
    """Test if kleio server url is available"""
    KLEIO_URL = kserver.get_url()
    assert KLEIO_URL is not None


def test_make_token():
    """Test if a token is generated"""
    assert KleioServer.make_token() is not None


def test_stop_kleio_server(setup):
    kserver: KleioServer = setup
    kome: str = kserver.get_kleio_home()

    """Test if kleio server is stopped"""

    kserver.stop()
    assert KleioServer.get_server(kome) is None
    kleio_home = f"{TEST_DIR}/timelink-home"
    assert KleioServer.get_server(kleio_home) is None
    kserver.start(kleio_home=kleio_home, update=False)
    # wait for server to start
    import time

    time.sleep(1)
    assert KleioServer.is_server_running(kleio_home=kleio_home) is True


def test_generate_limited_token(setup):
    kserver: KleioServer = setup
    """Generate a token with limited privileges"""
    user: str = "limited_user"
    info: TokenInfo = TokenInfo(
        **{
            "comment": "An user that has no privileges, used to test authorization errors",
            "api": [],
            "structures": "projects/test-project/structures",
            "sources": "projects/test-project/sources",
        }
    )

    try:  # we try to invalidade first
        kserver.invalidate_user(user)
    except Exception:
        pass

    limited_token = kserver.generate_token(user, info)
    assert limited_token is not None

    try:
        kserver.translation_status("", recurse=False, token=limited_token)
        raise AssertionError("This should not happen, user has no privileges")
    except KleioServerForbidenException:
        assert True


def test_generate_normal_token(setup):
    """Generate a token for normal user"""
    user: str = "normal_user"
    info: TokenInfo = TokenInfo(
        **{
            "comment": "An user able to translate, upload and "
            "delete files, and also create and "
            "remove directories, in specific sub-directories in kleio-home",
            "api": [
                "sources",
                "kleioset",
                "files",
                "structures",
                "translations",
                "upload",
                "delete",
                "mkdir",
                "rmdir",
            ],
            "structures": "projects/test-project/structures",
            "sources": "projects/test-project/sources",
        }
    )

    kserver: KleioServer = setup

    try:
        invalidated = kserver.invalidate_user(user)
        if invalidated is None:
            print("User not invalidated")
    except Exception:
        pass

    KLEIO_NORMAL_TOKEN = kserver.generate_token(user, info)
    assert KLEIO_NORMAL_TOKEN is not None


def test_get_kserver_home():
    """Test get the mapped kleio home from running server"""

    kleio_home = kleio_server.get_kserver_home()
    assert kleio_home is not None


def test_translations_get(setup):
    """Test if translations are retrieved"""
    path: str = "projects/test-project/kleio/reference_sources/linked_data"
    recurse: str = "yes"
    status: str = None

    kserver: KleioServer = setup
    translations = kserver.translation_status(path,
                                              recurse=recurse,
                                              status=status)
    assert len(translations) > 0

    kfile: KleioFile
    print()
    for kfile in translations:
        print(
            f"{kfile.status} {kfile.path} M:{kfile.modified} "
            f"T:{kfile.translated} E:{kfile.errors} W:{kfile.warnings}"
        )
    print()


def test_translations_translate(setup):
    """Test if translations are translated"""
    path: str = "projects/test-project/kleio/reference_sources/linked_data"
    recurse: str = "yes"
    spawn: str = "no"

    kserver: KleioServer = setup
    translations = kserver.translate(path, recurse, spawn)
    assert translations is not None


def test_translations_processing(setup):
    """Test translations in process"""
    path: str = "projects/test_project/"
    recurse: str = "yes"
    status: str = "P"

    kserver: KleioServer = setup
    translations = kserver.translation_status(path, recurse, status)
    assert len(translations) is not None

    kfile: KleioFile
    for kfile in translations:
        print(
            f"{kfile.status} {kfile.path} M:{kfile.modified} "
            f"T:{kfile.translated} E:{kfile.errors} W:{kfile.warnings}"
        )


def test_translations_queued(setup):
    """Test translation queued"""
    path: str = "projects/test-project/"
    recurse: str = "yes"
    status: str = "Q"

    kserver: KleioServer = setup
    translations = kserver.translation_status(path, recurse, status)
    assert translations is not None

    kfile: KleioFile
    for kfile in translations:
        print(
            f"{kfile.status} {kfile.path} M:{kfile.modified} "
            f"T:{kfile.translated} E:{kfile.errors} W:{kfile.warnings}"
        )


def test_translations_clean(setup):
    """Test if translations results are deleted"""
    path: str = "projects/test-project/kleio/reference_sources/linked_data"
    recurse: str = "yes"

    kserver: KleioServer = setup

    queued = kserver.translation_status(path, recurse, "Q")
    while len(queued) > 0:
        print(f"Waiting for {len(queued)} queued translations to finish")
        import time

        time.sleep(1)
        queued = kserver.translation_status(path, recurse, "Q")

    processing = kserver.translation_status(path, recurse, "P")
    while len(processing) > 0:
        print(f"Waiting for {len(processing)} processing translations to finish")
        import time

        time.sleep(1)
        processing = kserver.translation_status(path, recurse, "P")

    translations = kserver.translation_clean(path, recurse)
    assert translations is not None


def test_sources_get(setup):
    """Test if sources are retrieved"""
    path: str = ""
    recurse: str = "yes"

    kserver: KleioServer = setup
    sources = kserver.get_sources(path, recurse)
    assert sources is not None


def test_homepage_get(setup):
    """Test if homepage is retrieved"""
    url = "https://timelink.uc.pt/kleio"
    ks = KleioServer.attach(url, '', '')
    home = ks.get_home_page()
    # extract lines with pattern "([A-Za-z_]+):(.*)"
    pattern = re.compile(r"([A-Za-z_\ ]*):(.*)")

    matches = pattern.findall(home)
    home_page_info = {}
    for key, value in matches:
        print(f"{key} = {value}")
        home_page_info[key] = value
    version = home_page_info['Version']
    assert version is not None


def test_start_kleio_server_env():
    """Test if kleio server is started with env variables"""
    khome = f"{TEST_DIR}/timelink-home"
    kadmin_token = "0123456789"
    kserver_port = 8089
    kserver_exposed_port = 8989
    kworkers = 5
    kleio_conf_dir = "/kleio-home/systems/conf/kleio"
    kleio_source_dir = "/kleio-home/projects/test-project/sources"
    kleio_structure_dir = "/kleio-home/projects/test-project/structures"
    kleio_default_stru = "/kleio-home/projects/test-project/structures/sources.str"
    kleio_token_db = "/kleio-home/systems/conf/kleio/tokens.db"
    kleio_debug = "true"

    ks = KleioServer.start(
        kleio_home=khome,
        kleio_admin_token=kadmin_token,
        kleio_server_port=kserver_port,
        kleio_external_port=kserver_exposed_port,
        kleio_server_workers=kworkers,
        kleio_idle_timeout=60,
        kleio_conf_dir=kleio_conf_dir,
        kleio_source_dir=kleio_source_dir,
        kleio_stru_dir=kleio_structure_dir,
        kleio_default_stru=kleio_default_stru,
        kleio_token_db=kleio_token_db,
        kleio_debug=kleio_debug,
        reuse=False,
        update=False)

    time.sleep(3)
    home = ks.get_home_page()
    # extract lines with pattern "([A-Za-z_]+):(.*)"
    pattern = re.compile(r"([A-Za-z_\ ]*):(.*)")

    matches = pattern.findall(home)
    home_page_info = {}
    for key, value in matches:
        print(f"{key} = {value}")
        home_page_info[key] = value
    assert home_page_info['Workers'].strip() == str(kworkers), "number of works do not match"
    ptoken = home_page_info['Kleio_admin_token'].strip()
    assert ptoken == kadmin_token[:5], "token does not match"



from timelink.kleio.groups.kgroup import KGroup
from timelink.kleio.groups.kperson import KPerson
from timelink.kleio.groups.kobject import KObject
from timelink.kleio.groups.ksource import KSource
from timelink.kleio.groups.kact import KAct


class KRelation(KGroup):
    """KRelation(type,value,destname,destination[,date=,obs=])

    A relation between historical entities.

    Relations have a type, a value, a date and a destination.
    The origin of the relation is the entity represented by the
    group that includes the relation.

    Elements:
        type: the type of the relation. A String
        value: the value of the relation. A string.
        origin: the id of the origin of the destination. A string
        destination: the id of the destination of the relation. A string.
        destname: the name of the destination of the relation. A string
        date: the date of relation. A string in Timelink format, optional.
        obs: a note on the attribute. A string optional.

    Kleio stru definition:
        part 	name=relation ;
                position=type,value,destname,destination;
                guaranteed=type,value,destname,destination ;
                also=obs,date
    """

    _name = "rel"
    _position = ["type", "value", "destname", "destination"]
    _guaranteed = ["type", "value", "destname", "destination"]
    _also = ["obs", "date", "origin"]
    _pom_class_id: str = "relation"

    def before_include(self, container_group):
        """Method called before a new krelation is included
         in a group with container_group.include(KGroup).

         It sets the origin of the relation to the id of the container group.

        Args:
            container_group (KGroup): the group that includes the relation

        Returns:
            True if the relation can be included in the group, False otherwise.

        """
        KGroup.before_include(self, container_group)
        self["origin"] = container_group.id
        return True

    def after_include(self, group):
        pass


KPerson.allow_as_part(KRelation)
KObject.allow_as_part(KRelation)
KAct.allow_as_part(KRelation)
KSource.allow_as_part(KRelation)
KGroup._rel_class = KRelation



# Import all the models, so that Base has them before being
# imported by Alembic
# pylint: disable=import-error

from timelink.mhk.models.base_class import Base  # noqa
from timelink.mhk.models.entity import Entity  # noqa
from timelink.mhk.models.pom_som_mapper import PomSomMapper, PomClassAttributes  # noqa
from timelink.mhk.models.attribute import Attribute  # noqa
from timelink.mhk.models.relation import Relation  # noqa
from timelink.mhk.models.person import Person  # noqa
from timelink.mhk.models.object import Object  # noqa
from timelink.mhk.models.act import Act  # noqa
from timelink.mhk.models.source import Source  # noqa



from typing import Any

from box import Box

from timelink.kleio.utilities import quote_long_text


class KElement:
    """Represents an Element in Kleio language.

     While *Groups* represent historical entities (people, objects, events)
     *Elements* encapsulate basic items of information (name, gender, date).

    The value of an Element can have three possible "aspects":

     1) "core":  the actual information for the element
     2) "original" (optional), the original wording when relevant
     3) "comment" (optional), a comment of the value.

     Example in Kleio notation::

         person$Joaquim Carvalho%Joachim Carvº#Family name added in the margin

     Can be generated by ::

         n = KElement('name','Joaquim Carvalho',original='Joachim Carvº',
                 comment='Family name added in the margin')
         id = KElement('id','p-jrc')
         person = KPerson(id=id,name=n)

     TODO: does not deal with multiple values in elements.
     Check KElement in MHK where core, comment and original are Vectors
     better to create a KAspect class.

    """

    name: str = None
    core: Any  # must have a str representation.
    comment: str = None
    original: str = None
    _element_class = None

    @property
    def element_class(self):
        """This return _element_class if existing or name if not."""
        if self._element_class is not None:
            return self._element_class
        else:
            return self.__class__.name

    @element_class.setter
    def element_class(self, value):
        self._element_class = value
        pass

    def __init__(
        self,
        name: str = None,
        core: Any = None,
        comment=None,
        original=None,
        element_class=None,
    ):
        """
        Args:
            name: name of the Element. If None then self._name is used.
            core: the core aspect of the Element. Must have __str__
                    or a tuple (core,comment,original). If a tuple
                    optional arguments are disregarded.
            comment: Optional; The comment aspect of the Element.
            original: Optional; The original aspect of the Element.
            element_class: Optional; in groups from kleio translations
                this is set by the translator from the the source=parameter.
                of the str file. If absent it is set here
                equal to the name of the element.
        """
        if name is not None:
            self.name = name

        self.core = None
        self.comment = None
        self.original = None

        if type(core) is tuple and len(core) == 3:
            self.core = core[0]
            self.comment = core[1]
            self.original = core[2]
        else:
            self.core = core
            if comment is not None:
                self.comment = comment
            if original is not None:
                self.original = original
        if element_class is not None:
            # if element_class is not None then check if exists and store
            kclass = self.__class__.get_class_for(element_class)
            if kclass is not None:
                self._element_class = kclass
            else:
                # if element_class does not exist create a new class
                self._element_class = self.__class__.extend(name)

    def __str__(self):
        return self.core

    def __int__(self):
        return int(str(self))

    @classmethod
    def extend(cls, name: str):
        """
        Creates a new KElement class that extends this one.
        This allows creating new KElement
        names for different languages and keep the behaviour of
        a specific element type (e.g. mapping "data" elements
        behave like "date" elements).

        When an element of a group is set with an atomic value
        or a tuple, a new KElement of the class with the same
        name of the element in the group is used to store the
        group.

        :param name:
        :return:
        """
        new_kelement = type(name, (cls,), {})
        new_kelement.name = name
        return new_kelement

    @classmethod
    def get_subclasses(cls):
        """Generator for subclasses of this KElement"""
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def all_subclasses(cls):
        """List of all the subclasses of this KElement"""
        return list(cls.get_subclasses())

    @classmethod
    def get_class_for(cls, name: str):
        """
        Search in KElement subclasses and return the one
        with the same element name as the argument.

        if more than one apply return the more specialized (longer __mro__)

        If not found return None

        :param name: name of an element
        :return: KElement or a subclass
        """
        search_list = sorted(
            [(a, len(a.__mro__)) for a in cls.all_subclasses()], key=lambda mro: -mro[1]
        )
        for eclass, _mro in search_list:
            if eclass.name == name:
                return eclass
        return None

    @classmethod
    def inherits_from(cls):
        return [ke for ke in cls.__mro__ if ke is not object and ke is not None]

    @classmethod
    def get_classes_for(cls, name: str):
        """Search in Element subclass and return all that
        have the same element name as the argument

        see KElement.get_class_for(name) for getting the more specialized"""
        return [sc for sc in KElement.all_subclasses() if sc.name == name]

    def inherited_names(self):
        """Return the list of names in the KElement subclasses
        this element inherits from including its own"""
        return [
            ke.name
            for ke in self.__class__.__mro__
            if ke is not object and ke.name is not None
        ]

    def extends(self, name: str):
        """True if name is in the the classes this element inherits from"""
        return name in self.inherited_names()

    def is_empty(self):
        """True if all aspects of the element are None or empty string"""
        e = [
            x for x in [self.core, self.comment, self.original] if x is None or x == ""
        ]
        if len(e) == 3:
            return True
        else:
            return False

    def to_tuple(self):
        """Return Element as a tuple (core,comment,original)"""
        return self.core, self.comment, self.original

    def to_kleio(self, name=True, force_show=False):
        """
        Return element as a kleio string: element=core#comment%original
        To avoid rendering the element name set name=False
        :param name: if True(default) prefix value with self.name=
        :return: a valid Kleio representation of this element value
        """
        if self.extends("invisible_") and force_show is False:
            return ""
        if self.is_empty():
            return ""
        c = self.core
        cc = self.comment
        o = self.original
        if c is None or c == "":
            c = ""
        else:
            c = quote_long_text(str(c))
        if cc is None or cc == "":
            cc = ""
        else:
            cc = "#" + quote_long_text(str(cc))
        if o is None or o == "":
            o = ""
        else:
            o = "%" + quote_long_text(str(o))
        kleio = c + cc + o
        if name:
            kleio = f"{self.name}={kleio}"
        return kleio

    def to_dict(self, name=False):
        """Return Element as a dict {core:_, comment:_, original:_}
        add name=True to add name to dictionary:
        {name:_, core:_, comment:_, original:_}"""
        if name:
            return {
                "name": self.name,
                "core": self.core,
                "comment": self.comment,
                "original": self.original,
            }
        else:
            return {
                "core": self.core,
                "comment": self.comment,
                "original": self.original,
            }

    def to_dots(self):
        return Box(self.to_dict())


# use this for mixins to mark this should show in Kleio
# still can override with KElement.to_kleio(force_show=True
KleioNoShow = KElement.extend("invisible_")


# Default KElement classes gacto2.str
# element name=id; identification=yes
# element name=type
# element name=loc
# element name=obs
# element name=ref
# element name=value
# element name=origin
# element name=destination
# element name=entity
# element name=same_as
# element name=xsame_as
# element name=name
# element name=sex
# element name=destname
# element name=destination
# element name=summary;
# element name=description;
# element name=replace;


class KDate(KElement):
    name = "date"

    def __init__(
        self,
        date: Any = None,
        core=None,
        comment=None,
        original=None,
        element_class=None,
    ):
        if core is not None:
            date = core  # to allow core setting in generic code
        super().__init__(self.name, date, comment, original, element_class)


class KDay(KElement):
    name = "day"

    def __init__(
        self,
        day: Any = None,
        core=None,
        comment=None,
        original=None,
        element_class=None,
    ):
        if core is not None:
            day = core  # to allow core setting in generic code
        if day is None:
            day = 0

        super().__init__(self.name, day, comment, original, element_class)

        if type(self.core) is str:
            self.core = int(self.core)
        if self.core != 0 and (self.core < 1 or self.core > 31):
            raise ValueError("Day value must be between 1 and 31")


class KMonth(KElement):
    name = "month"

    def __init__(
        self,
        month: Any = None,
        core=None,
        comment=None,
        original=None,
        element_class=None,
    ):
        if core is not None:
            month = core  # to allow core

        if month is None:
            month = 0

        super().__init__(self.name, month, comment, original, element_class)

        self.core = int(self.core)

        if type(self.core) is str:
            self.core = int(self.core)
        if self.core != 0 and (self.core < 1 or self.core > 12):
            raise ValueError("Month value must be between 1 and 12")


class KYear(KElement):
    """
    Represents a year.

    To have value checking do KYear.set_limits((lower,upper)) !!!!

    """

    name = "year"

    def __init__(self, year: Any = None, core=None, comment=None, original=None):
        if core is not None:
            year = core

        if year is None:
            year = 0

        super().__init__(self.name, year, comment, original)

        try:
            self.core = int(self.core)
        except ValueError:
            # warn of value error
            # this is a hack because many sources do not use the second position as year, and
            # all sort of data gets there
            self.core = core


class KType(KElement):
    """
    Represents a type of object or abstraction
    """

    name = "type"


class KValue(KElement):
    """
    Represents a general purpose value
    """

    name = "value"


class KId(KElement):
    """
    Represents an unique id for a group.

    """

    name = "id"


class KEntityInAttribute(KId, KleioNoShow):
    name = "entity"


class KOriginInRel(KId, KleioNoShow):
    name = "origin"


class KReplace(KElement):
    """
    Represents the id of a group to be replaced.

    Example:

        source$new-id/replace=old-id
    """

    name = "replace"


class KSameAs(KElement):
    r"""
    Represents the id of a group that describes the
    same real world entity has the one with this element.

    Used in the same file. Translators should check if
    the id corresponds to a group in the same file and
    file an error otherwise

    Examples:

    person$Bob Dylan/id=bob-dylan

    person$Robert Allan Zimmerman/sameas=bob-dylan

    """

    name = "same_as"


class KXSameAs(KElement):
    """
    Same meaning as KSameAs used when id is not
    in the file.

    The difference between KSameAs and KXSameAs is
    just for error checking during translation.
    Translators will raise error if a KSameAs id
    is not found in the same file, but only a warning
    for KXSameAs.

    """

    name = "xsame_as"


class KName(KElement):
    """
    Name of person
    """

    name = "name"


class KSex(KElement):
    """
    male / female ...
    """

    name = "sex"


class KDescription(KElement):
    """
    Similar to name, for objects
    """

    name = "description"


class KObs(KElement):
    """
    Element for "obs" normally observations or notes
    """

    name = "obs"


class KStructure(KElement):
    """
    Element for structure name in sources
    """

    name = "structure"


class KLoc(KElement):
    """
    Element for location (in a document, e.g. page)  in some groups
    """

    name = "loc"


class KRef(KElement):
    """
    Element for reference number (e.g. a call number in an archive)
    """

    name = "ref"


class KDestName(KElement):
    """
    Element for destination names in relations
    """

    name = "destname"


class KDestId(KElement):
    """
    Element for destination id in relations
    """

    name = "destination"


class KSummary(KElement):
    """
    Element for summaries (long texts)
    """

    name = "summary"


class KReplaceSourceId(KElement):
    """
    Element for id of source being replaced in source groups
    """

    name = "replace"


# Automatic elements, generated during the translation process


class KKleiofile(KElement):
    """
    Element for the name of the file from which the source was imported
    """

    name = "kleiofile"


class KGroupName(KElement):
    """
    Element for groupname of a group
    """

    name = "groupname"


class KClass(KElement):
    """
    Element for the class of a group
    """

    name = "class"


class KLevel(KElement):
    """
    Element for nesting level in imported groups
    """

    name = "level"


class KLine(KElement):
    """
    Element for line number in imported groups
    """

    name = "line"


class KOrder(KElement):
    """
    Element for group order in imported groups
    """

    name = "order"


class KInside(KElement):
    """
    Element for id of the enclosing group
    """

    name = "inside"


class KUndef(KElement):
    """
    Element of undefined class (generated by the kleio translator)

    This is generated when the translator finds an element for which
    the is no mapping POM_SOM in any class. An example if the elements
    for day,month,year that are used for input but are stored as a single
    value YYYYMMDD and so are not present in the mapping information for
    database storage.
    """

    name = "undef"



# Pydantic schema
from typing import Annotated, List, Optional
from datetime import datetime
from pydantic import BaseModel, ConfigDict, SkipValidation

from timelink.api.database import TimelinkDatabase
from timelink.kleio.kleio_server import KleioServer


class UserPropertySchema(BaseModel):
    name: str
    value: str

    model_config = ConfigDict(from_attributes=True)


class UserProjectSchema(BaseModel):

    user_id: Optional[int] = None
    project_id: Optional[int] = None
    project_name: Optional[str] = None
    project_description: Optional[str] = None
    access_level: Optional[str] = None

    # ORM mode
    model_config = ConfigDict(from_attributes=True)


class UserSchema(BaseModel):
    """User schema.

    Fields:
    - name: str
    - email: str
    - nickname: Optional[str]
    - disabled: Optional[bool]
    - properties: Optional[List[UserPropertySchema]]"""
    name: str
    email: str
    nickname: Optional[str]
    disabled: Optional[bool]
    properties: Optional[List[UserPropertySchema]]
    projects: Optional[List["UserProjectSchema"]]  # noqa Flake8: F821
    current_project_name: Optional[str] = None
    current_project_db: Annotated[TimelinkDatabase, SkipValidation] = None
    current_project_kleio_server: Annotated[KleioServer, SkipValidation] = None
    created: Optional[datetime] = None
    updated: Optional[datetime] = None

    def is_admin(self, admin_property: str = "permission:timelink:admin"):
        for prop in self.properties:
            if prop.name == admin_property and prop.value in ["true", "yes"]:
                return True
        return False

    # ORM mode
    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)



from typing import Optional

# pylint: disable=import-error
from sqlalchemy import String, ForeignKey
from sqlalchemy.orm import mapped_column, Mapped

from .entity import Entity
from timelink.kleio.utilities import quote_long_text


class Person(Entity):
    """Represents a person in a historical source"""

    __tablename__ = "persons"

    id: Mapped[str] = mapped_column(
        String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True
    )
    name: Mapped[Optional[str]] = mapped_column(String, index=True)
    sex: Mapped[Optional[str]] = mapped_column(String(1))
    obs: Mapped[Optional[str]] = mapped_column(String)

    __mapper_args__ = {"polymorphic_identity": "person"}

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Person(id={sr}, "
            f'name="{self.name}", '
            f'sex="{self.sex}", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        r = (
            f"{self.groupname}${quote_long_text(self.name)}/"
            f"{quote_long_text(self.sex)}/id={quote_long_text(self.id)}"
        )
        if self.obs is not None:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r


def get_person(id: str = None, db=None, sql_echo: bool = False) -> Person:
    """
    Fetch a person from the database
    """
    if id is None:
        raise (Exception("Error, id needed"))
    p: Person = db.session().get(Person, id)
    return p


def pperson(id: str):
    """Prints a person in kleio notation"""
    print(get_person(id=id).to_kleio())



import os

# pylint: disable=import-error

import pytest

# Session is shared by all tests
from tests import Session
from tests import skip_on_travis, conn_string
from timelink.kleio.groups import KElement, KGroup, KSource, KAct, KPerson
from timelink.mhk.models import base  # pylint: disable=unused-import. # noqa: F401
from timelink.mhk.models.base_class import Base
from timelink.mhk.models.db import TimelinkMHK
from timelink.mhk.models.entity import Entity  # noqa
from timelink.mhk.models.pom_som_mapper import PomSomMapper
from timelink.mhk.models.source import Source
from timelink.mhk.utilities import is_mhk_installed

pytestmark = skip_on_travis

if not is_mhk_installed():
    pytest.skip("skipping MHK tests (MHK not present)", allow_module_level=True)


@pytest.fixture(scope="module")
def dbsystem():
    db = TimelinkMHK(conn_string)
    Session.configure(bind=db.get_engine())
    yield db
    with Session() as session:
        db.drop_db(session)


@pytest.fixture
def kgroup_person_attr_rel() -> KSource:
    """Returns a group wtih attr and rel"""
    ks = KSource('s2', type='test', loc='auc', ref='attr-rel', obs="""
        This is a multiline obs with special characteres like /,= and $
        """)
    ka1 = KAct('a2-1', 'test-act', date='2021-07-16',
               day=16, month=7, year=2021,
               loc='macau', ref='p.1', obs='Test Act')
    ks.include(ka1)
    p1 = KPerson('Joaquim', 'm', 'p01-2', obs="Living in Macau/China")
    p1.attr('residencia', 'Macau', date='2021-12-11')
    mobs = """
    Moving from Portugal
    to Macau in 2022
    """
    p2 = KPerson('Margarida', 'f', 'p02-2', obs=mobs)
    p2.attr('residencia', 'Trouxemil', date='2020-10-18')
    p1.rel('parentesco', 'marido', p2.name, p2.id,  # pylint: disable=no-member
           date='2006-01-4', obs='Ilha Terceira')
    ka1.include(p2)
    ka1.include(p1)
    return ks


@pytest.fixture
def kgroup_nested() -> KSource:
    """Returns a nested structure"""
    ks = KSource('s1', type='test', loc='auc', ref='alumni', obs='Nested')
    ka1 = KAct('a1', 'test-act', date='2021-07-16',
               day=16, month=7, year=2021,
               loc='auc', ref='p.1', obs='Test Act')
    ks.include(ka1)
    ka2 = KAct('a2', 'test-act', date='2021-07-17',
               day=17, month=7, year=2021,
               loc='auc', ref='p.2', obs='Test Act')

    ks.include(ka2)
    p1 = KPerson('Joaquim', 'm', 'p01')
    p1.attr('residencia', 'Macau', date='2021-12-11')
    p2 = KPerson('Margarida', 'f', 'p02')
    p2.attr('residencia', 'Trouxemil', date='2020-10-18')
    p3 = KPerson('Pedro', 'm', 'p03')
    p3.attr("residencia", "Coimbra", date='2021-10-21')

    ka1.include(p3)
    p4 = KPerson("Maria", "f", "p04")
    p5 = KPerson("Manuel", "m", "p05")
    p6 = KPerson("João", "m", "p06")
    ka2.include(p4)
    ka2.include(p5)
    ka2.include(p6)
    return ks


@skip_on_travis
def test_succeed_if_not_in_travis():
    assert os.getenv("TRAVIS") != 'true'


def test_create_db(dbsystem):
    # Database is created and initialized in the fixture
    metadata = Base.metadata  # pylint: disable=no-member
    tables = list(metadata.tables.keys())
    assert len(tables) > 0, "tables where not created"


@skip_on_travis
def test_entity_contains(dbsystem):
    ent1: Entity = Entity(id='e001', groupname='entity')
    ent1.pom_class = 'entity'
    ent1.line = 1
    ent1.level = 1
    ent1.sequence = 1

    ent2: Entity = Entity(id="e002")
    ent2.groupname = "entity"
    ent2.pom_class = "entity"
    ent2.line = 2
    ent2.level = 2
    ent2.sequence = 2
    ent1.contains.append(ent2)

    ent3: Entity = Entity(id="e003")
    ent3.groupname = "entity"
    ent3.pom_class = "entity"
    ent3.line = 2
    ent3.level = 2
    ent3.sequence = 2
    ent1.contains.append(ent3)

    ent4: Entity = Entity(id="e004")
    ent4.groupname = "entity"
    ent4.pom_class = "entity"
    ent4.line = 4
    ent4.level = 4
    ent4.sequence = 4
    ent2.contains.append(ent4)

    with Session() as session:
        session.add(ent1)
        session.add(ent2)
        session.add(ent3)
        session.commit()
        ent4 = ent2.contained_by  # pylint: disable=no-member
        assert ent4 is ent1
        ent5 = ent1.contains[1]
        assert ent5 is ent3
        session.delete(ent1)  # should delete ent2, ent3 and ent4
        session.commit()
        deleted = session.get(Entity, "e004")
        assert deleted is None, "Should have deleted orphan"


def test_ensure_mapping(dbsystem):
    with Session() as session:
        pom_classes = PomSomMapper.get_pom_classes(session=session)
        pom_ids = PomSomMapper.get_pom_class_ids(session=session)
        pom_tables = [pom_class.table_name for pom_class in pom_classes]
        pom_class: PomSomMapper
        for pom_class in pom_classes:
            pom_class.ensure_mapping(session=session)
            # print(repr(pom_class))
            # print(pom_class)
        orm_mapped_classes = Entity.get_som_mapper_to_orm_as_dict()
        non_mapped = set(pom_ids) - set(orm_mapped_classes.keys())
        assert len(non_mapped) == 0, "Not all classes are mapped to ORM"
        orm_mapped_tables = Entity.get_tables_to_orm_as_dict()
        tables_not_mapped = set(pom_tables) - set(orm_mapped_tables)
        assert len(
            tables_not_mapped) == 0, "Not all class tables are mapped in ORM"
        db_tables = dbsystem.table_names()
        tables_not_in_db = set(orm_mapped_tables) - set(db_tables)
        assert len(
            tables_not_in_db) == 0, \
            "Not all mapped tables were created in the db"
        session.close()


def test_insert_entities_nested_groups(dbsystem, kgroup_nested):
    ks = kgroup_nested
    source_id = ks.get_id()
    with Session() as session:
        PomSomMapper.store_KGroup(ks, session)
        session.commit()
        source_from_db: Source = Entity.get_entity(source_id, session)
        assert source_from_db.id == ks.get_id()
        assert repr(source_from_db)
        assert str(source_from_db)
        acts = source_from_db.contains
        act = acts[0]
        ract = repr(act)
        sact = str(act)
        assert ract
        assert sact
        person = act.contains[0]
        assert repr(person)
        assert str(person)


def test_insert_entities_attr_rel(dbsystem, kgroup_person_attr_rel):
    ks = kgroup_person_attr_rel
    source_id = ks.get_id()
    with Session() as session:
        PomSomMapper.store_KGroup(ks, session)
        session.commit()
        source_from_db = Entity.get_entity(source_id, session)
        assert source_from_db.id == ks.get_id()


def test_quote_and_long_test(dbsystem, kgroup_person_attr_rel):
    ks = kgroup_person_attr_rel
    kleio = ks.to_kleio()
    assert '"""' in kleio, "Bad handling of long text"


def test_insert_delete_previous_source(dbsystem, kgroup_nested):
    ks = kgroup_nested
    source_id = ks.get_id()
    with Session() as session:
        print(
            f"""

        =================
        Inserting a nested group

        {ks.to_kleio()}
        """
        )
        PomSomMapper.store_KGroup(ks, session)
        session.commit()
        ks2 = KSource('s1', type='test',
                      loc='auc',
                      ref='alumni',
                      obs='Empty should delete previous one')

        print(
            """
        ================
        Reinserting group with same id
        ===============
        """
        )
        PomSomMapper.store_KGroup(ks2, session)
        source_from_db: Entity = Entity.get_entity(source_id, session)
        contains = source_from_db.contains
        assert len(contains) == 0, "import did not delete contained entities"


def test_store_KGroup_1(dbsystem):
    kfonte: KGroup = KGroup.extend('fonte',
                                   position=['id', 'data'],
                                   also=['tipo',
                                         'ano',
                                         'obs',
                                         'substitui'])
    afonte = kfonte('f001',
                    data=KElement('data', '2021-12-02', element_class='date'),
                    tipo=KElement('tipo', 'teste', element_class='type'),
                    obs="First group stored through ORM with non core group")
    with Session() as session:
        session.commit()
        afonte._pom_class_id = 'source'  # need a set_mapper method
        # this converts a group to an database entity
        afonte_entity = PomSomMapper.kgroup_to_entity(afonte, session)
        assert afonte_entity is not None
        PomSomMapper.store_KGroup(afonte, session)
        session.commit()
        source_orm = Entity.get_orm_for_pom_class(afonte.pom_class_id)
        same_fonte = session.get(source_orm, afonte.id.core)
        assert str(afonte.obs) == str(same_fonte.obs)  # pylint: disable=no-member
        session.close()



""" CRUD operations for the timelink API. """

from datetime import datetime
from sqlalchemy.orm import Session  # pylint: disable=import-error
from timelink.api import models
from timelink.api.schemas import EntityAttrRelSchema


def get_syspar(db: Session, q: list[str] | None = None):
    """Get system parameters
    Args:
        db: database session
        q: parameter name(s); if empty, return all parameters
    Returns:
        SysPar object
    """
    if q:
        if isinstance(q, str):
            q = [q]
        return db.query(models.SysPar).filter(models.SysPar.pname.in_(q)).all()
    return db.query(models.SysPar).all()


def set_syspar(
    db: Session, syspar: models.SysParSchema
):  # pylint: disable=invalid-name
    """Set system parameters
    Args:
        db: database session
        syspar: SysPar object
    Returns:
        SysPar object
    """
    db_syspar = get_syspar(db, syspar.pname)
    if db_syspar:
        db_syspar.pvalue = syspar.pvalue
        db_syspar.ptype = syspar.ptype
        db_syspar.obs = syspar.obs
    else:
        db_syspar = models.SysPar(
            pname=syspar.pname, pvalue=syspar.pvalue, ptype=syspar.ptype, obs=syspar.obs
        )
        db.add(db_syspar)
    db.commit()
    db.refresh(db_syspar)
    return db_syspar


def get_syslog(
    db: Session, nlogs: int
) -> list[models.SysLog]:  # pylint: disable=invalid-name
    """Get last n system logs last one first
    Args:
        db: database session
        nlogs: sequence number
    Returns:
        List of SysLog objects
    """
    return db.query(models.SysLog).order_by(models.SysLog.seq.desc()).limit(nlogs).all()


def get_syslog_by_time(
    db: Session,  # pylint: disable=invalid-name
    start_time: datetime,
    end_time: datetime,
) -> list[models.SysLog]:
    """Get system logs between start_time and end_time
    Args:
        db: database session
        start_time: start time
        end_time: end time
    Returns:
        List of SysLog objects
    """
    return (
        db.query(models.SysLog)
        .filter(models.SysLog.time >= start_time)
        .filter(models.system.SysLog.time <= end_time)
        .all()
    )


def set_syslog(
    db: Session, log: models.system.SysLogCreateSchema
) -> models.system.SysLog:  # pylint: disable=invalid-name
    """Set system log
    Args:
        db: database session
        log: SysLogCreateSchema object with level, origin and message
    Returns:
        SysLog object
    """
    db_syslog = models.SysLog(origin=log.origin, message=log.message, level=log.level)
    db.add(db_syslog)
    db.commit()
    db.refresh(db_syslog)
    return db_syslog


def get(db: Session, id: str) -> EntityAttrRelSchema:  # pylint: disable=invalid-name
    """Get entity by id
    Args:
        db: database session
        id: entity id
    Returns:
        Entity object
    """
    entity = models.Entity.get_entity(id, db)
    # get the columns of this entity
    pentity = EntityAttrRelSchema.from_orm(entity)
    # get the relations of this entity

    # TODO return the entity as a dictionary with rels in and out
    #       and contains

    return pentity



"""
Create a dataframe with the values of an attribute
"""

import pandas as pd

from sqlalchemy import select, func, and_, desc

from timelink.api.database import TimelinkDatabase
from timelink.api.models import Attribute


def attribute_values(
    attr_type,
    db: TimelinkDatabase = None,
    dates_between=None,
    sql_echo=False,
):
    """Return the vocabulary of an attribute

    The returned dataframe has a row for each unique value
    a 'count' with the number of different entities, and
    the the first and last date for that row

    Args:
        attr_type = attribute type to search for
        db = database connection to use, either db or session must be specified
        dates_between = tuple with two dates in format yyyy-mm-dd
        sql_echo = if true will print the sql statement


    To filter by dates: dates_in = (from_date,to_date)
    with dates in format yyyy-mm-dd
    will return attributes with
    from_date < date < to_date

    """

    #  We try to use an existing connection and table introspection
    # to avoid extra parameters and going to database too much
    dbsystem: TimelinkDatabase = None
    if db is not None:  # try if we have a db connection in the parameters
        dbsystem = db
    else:
        raise Exception(
            "No database connection specified, must set up a database"
            " connection before or specify previously openned database"
            " with db="
        )

    attr_table = Attribute.__table__

    if dates_between is not None:
        first_date, last_date = dates_between
        stmt = (
            select(
                attr_table.c.the_value.label("value"),
                func.count(attr_table.c.entity.distinct()).label("count"),
                func.min(attr_table.c.the_date).label("date_min"),
                func.max(attr_table.c.the_date).label("date_max"),
            )
            .where(
                and_(
                    attr_table.c.the_type == attr_type,
                    attr_table.c.the_date > first_date.strip("-"),
                    attr_table.c.the_date < last_date.strip("-"),
                )
            )
            .group_by("the_value")
            .order_by(desc("count"))
        )
    else:
        stmt = (
            select(
                attr_table.c.the_value.label("value"),
                func.count(attr_table.c.entity.distinct()).label("count"),
                func.min(attr_table.c.the_date).label("date_min"),
                func.max(attr_table.c.the_date).label("date_max"),
            )
            .where(attr_table.c.the_type == attr_type)
            .group_by("the_value")
            .order_by(desc("count"))
        )

    if sql_echo:
        print(stmt)

    with dbsystem.session() as session:
        records = session.execute(stmt)
    df = pd.DataFrame.from_records(
        records, index=["value"], columns=["value", "count", "date_min", "date_max"]
    )

    return df



# pylint: disable=import-error
from pathlib import Path

import pytest
from sqlalchemy.orm.session import Session

from tests import skip_on_travis, mhk_absent, conn_string, TEST_DIR
from timelink.kleio.importer import import_from_xml
from timelink.mhk.models import base  # pylint: disable=unused-import. # noqa: F401
from timelink.mhk.models.base import Person  # noqa pylint: disable=unused-import
from timelink.mhk.models.db import TimelinkMHK
from timelink.mhk.models.entity import Entity
from timelink.mhk.utilities import is_mhk_installed

# https://docs.pytest.org/en/latest/how-to/skipping.html

if not is_mhk_installed():
    pytest.skip("skipping MHK tests (MHK not present)", allow_module_level=True)

pytestmark = skip_on_travis


@pytest.fixture(scope="module")
def dbsystem():
    database = TimelinkMHK(conn_string)
    db = database.session()
    try:
        yield db
    finally:
        database.drop_db(db)
        db.close()


@skip_on_travis
@mhk_absent
def test_import_xml(dbsystem: Session):
    file: Path = Path(TEST_DIR, "xml_data/b1685.xml")
    session = dbsystem
    stats = import_from_xml(
        file, session, options={"return_stats": True, "mode": "MHK"}
    )
    sfile: Path = stats["file"]
    assert sfile.name == file.name
    domingos_vargas = session.get(Person, "b1685.33-per6")
    assert domingos_vargas is not None, "could not get a person from file"
    kleio = domingos_vargas.to_kleio()
    assert len(kleio) > 0


@skip_on_travis
@mhk_absent
def test_import_with_custom_mapping(dbsystem: Session):
    file = Path(TEST_DIR, "xml_data/dev1692.xml")
    session = dbsystem

    stats = import_from_xml(
        file, session, options={"return_stats": True, "mode": "MHK"}
    )
    sfile = stats["file"]
    assert "dev1692" in sfile.name
    caso = session.get(Entity, "c1692-antonio-cordeiro-alcouc")
    caso_kleio = caso.to_kleio()
    assert len(caso_kleio) > 0
    assert caso is not None, "could not get an entity from special mapping"
    caso_kleio = caso.to_kleio()
    assert len(caso_kleio) > 0
    test = session.get(Entity, "dev1692-per5")
    assert test is not None
    test_kleio = test.to_kleio()
    assert len(test_kleio) > 0


@skip_on_travis
@mhk_absent
def test_import_with_many(dbsystem: Session):
    file = Path(TEST_DIR, "xml_data/test-auc-alunos-264605-A-140337-140771.xml")
    session = dbsystem
    stats = import_from_xml(
        file, session, options={"return_stats": True, "mode": "MHK"}
    )
    sfile = stats["file"]
    assert "alunos" in sfile.name
    estudante = session.get(Entity, "140771")
    kleio = estudante.to_kleio()
    assert len(kleio) > 0
    assert estudante is not None, (
        "could not get an entity from big import" @ skip_on_travis
    )  # noqa


@skip_on_travis
@mhk_absent
def test_import_git_hub(dbsystem: Session):
    file = "https://raw.githubusercontent.com/time-link/timelink-py/f76007cb7b98b39b22be8b70b3b2a62e7ae0c12f/tests/xml_data/b1685.xml"  # noqa
    session = dbsystem
    stats = import_from_xml(
        file, session, options={"return_stats": True, "mode": "MHK"}
    )

    domingos_vargas = session.get(Person, "b1685.33-per6")
    assert domingos_vargas is not None, "could not get a person from file"
    kleio = domingos_vargas.to_kleio()
    assert len(kleio) > 0
    sfile: Path = stats["file"]
    assert "b1685.xml" in sfile
    pass



"""
Test models without requiring a db connection

"""

from timelink.api.models import base  # noqa
from timelink.api.models.entity import Entity  # noqa


def test_entity_subclasses():
    scl = list(Entity.get_subclasses())
    sc1 = len(scl)

    class SubEntity(Entity):
        # polymorphic_on = "type"

        __mapper_args__ = {
            "polymorphic_identity": "subentity",
            "inherit_condition": id == Entity.id,
        }
        pass

    scl2 = list(Entity.get_subclasses())
    sc2 = len(scl2)
    assert sc2 == sc1 + 1, "wrong direct subclasses of Entity"

    class SubSubEntity(SubEntity):
        __mapper_args__ = {
            "polymorphic_identity": "subsubentity",
            "inherit_condition": id == Entity.id,
        }
        pass

    scl3 = list(Entity.get_subclasses())
    sc3 = len(scl3)
    assert sc3 == sc2 + 1, "wrong recursive subclasses of Entity"



import json
import textwrap
import warnings
from os import linesep as nl
from typing import Type, List, Union, Tuple, Dict

from box import Box

from timelink.kleio.groups.kelement import KElement


class KGroup:
    """
    KGroup(*positional_elements ,**more_elements)

    Abstract Kleio Group.

    To define a Kleio Group extend this class and set default value for _name,
    or use extend(name,position, guaranteed, also).

    Define allowed elements in the default values for _position, _guaranteed,
    _also (list of strings) or call allowed_as_element .

    Use _part to list allowed enclosed groups, or call allow_as_part

    For an example see timelink.kleio.groups.KPerson

    """

    # Class scoped list of reserved element names for system use
    _builtin_elements = [
        "line",
        "level",
        "order",
        "inside",
        "groupname",
        "pom_class_id",
        "id",
    ]

    id = None
    _name: str = "kgroup"
    _position: list = []  # list of positional elements
    _guaranteed: list = []  # list of required elements
    _also: list = []  # list of optional elements
    _part: list = []  # allowed sub groups

    _extends: Type["KGroup"]  # name of group this is based on
    _pom_class_id: str = "entity"  # Id of PomSom mapper for this group
    _element_check = True  # if true validates element assignment
    _elementsd: Type[Dict[str, KElement]]  # Current elements
    _inside: Type["KGroup"]  # group that includes this

    # The following fields are generated during the translation process.
    # If the Kleio groups are generated programatically (for instance when
    # importing from csv files or from databases) then the cls.include method
    # will update these values
    _line: int  # line in the source file
    _level: int  # nesting always +1  than enclosing group
    _order: int  # sequential number of this group in the original source
    # class scoped counters to ensure proper numbering across different groups
    _global_sequence: int = 1  # global sequence count
    _global_line: int = 1  # global line count

    # we need the classes for attributes and relations, used by the
    # attr and rel methods. Since those class are subclasses of this one
    # they need to set this latter. Also allows for differet classes to
    # be used by attr and rel if necessary
    _attr_class = None  # class used by the attr method (set by KAttribute)
    _rel_class = None  # class used by the rel method (set by KRelation)

    # TODO to_kleio_str generates the definition of a group
    #       for a kleio str file. recurse=yes
    #       collects included groups and generates for those also.

    @property
    def kname(self):
        """The kleio name of this group, used in the to_kleio() method"""
        return self.unpack_from_kelement(self._name)

    @kname.setter
    def kname(self, value):
        self._name = self.pack_as_kelement("kname", value, element_class="name")

    @property
    def inside(self):
        return self._inside

    def get_container_id(self):
        if self._inside is None:
            gid = None
        elif KGroup.is_kgroup(self._inside):
            gid = self.unpack_from_kelement(self.inside.id)
        return gid

    @inside.setter
    def inside(self, value):
        if KGroup.is_kgroup(value):
            self._inside = value
        else:
            self._inside = KGroup(id=str(value), element_check=False)

    @property
    def line(self):
        return self.unpack_from_kelement(self._line)

    @line.setter
    def line(self, value):
        self._line = self.pack_as_kelement("line", value)

    @property
    def level(self):
        return self.unpack_from_kelement(self._level)

    @level.setter
    def level(self, value):
        self._level = self.pack_as_kelement("level", value)

    @property
    def order(self):
        return self.unpack_from_kelement(self._order)

    @order.setter
    def order(self, value):
        self._order = self.pack_as_kelement("order", value)

    @property
    def pom_class_id(self):
        return self.unpack_from_kelement(self._pom_class_id)

    @pom_class_id.setter
    def pom_class_id(self, pcid):
        self._pom_class_id = self.pack_as_kelement("pom_class_id", pcid)

    @classmethod
    def extend(
        cls,
        name: str,
        position: Union[list, str, None] = None,
        guaranteed: Union[list, str, None] = None,
        also: Union[list, str, None] = None,
        part: Union[list, str, None] = None,
        synonyms: Union[Tuple[str, str], None] = None,
    ):
        """
        Create a new group definition by extending this one

        :param name:  name of the new group
        :param position: list of positional elements
        :param guaranteed: list of required elements
        :param also:    list of optional elements
        :param part: list of groups that can be included in this one
        :param synonyms: list of tuples defining synonym for elements
                         e,g, [("ano","year"),...]
        :return: new group class



        """
        new_group = type(name, (cls,), {})
        new_group._name = name
        # todo: k,v in kwargs if in cls set if not error
        if type(position) is str:
            position = [position]
        if type(guaranteed) is str:
            guaranteed = [guaranteed]
        if type(also) is str:
            also = [also]
        if position is not None:
            new_group._position = position
        else:
            new_group._position = list(cls._position)
        if guaranteed is not None:
            new_group._guaranteed = guaranteed
        else:
            new_group._guaranteed = list(cls._guaranteed)
        if also is not None:
            new_group._also = also
        else:
            new_group._also = list(cls._also)
        if part is not None:
            new_group._part = part
        else:
            new_group._part = list(cls._part)
        if synonyms is not None:
            try:
                for word, synonym in synonyms:
                    sym_class = KElement.get_class_for(synonym)
                    if sym_class is None:
                        msg = f"synonym error {synonym} is unkown"
                        raise ValueError(msg)
                    else:
                        # this creates a new KElement class
                        # named "word" that inherits from the KElement
                        # class of synonym
                        sym_class.extend(word)
            except ValueError:
                msg = "synonyms must be list of tuples"
                "e.g. [('ano','date').."
                raise (ValueError(msg))

        new_group._extends = cls

        return new_group

    @classmethod
    def get_subclasses(cls):
        """Generator for subclasses of this group"""
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def all_subclasses(cls):
        """List of all the subclasses of this group"""
        return list(cls.get_subclasses())

    @classmethod
    def is_kgroup(cls, g):
        """True g is an instance of a subclass of KGroup"""
        return "KGroup" in [c.__name__ for c in type(g).mro()]

    @classmethod
    def elements_allowed(cls) -> set:
        """Set of  Elements allowed in this Group"""
        return set(cls._guaranteed).union(set(cls._also)).union(cls._position)

    @classmethod
    def allow_as_element(
        cls, ename: Union[str, List[str]], guaranteed=False, also=True, position=None
    ):
        """
        Add element or list to list of allowed elements for this group.
        Optionally define if element(s) is positional,
        required (guaranteed) or optional
        :param ename:  name of element
        :param guaranteed: if True this is element is added to list of required elements.
        :param also: if True this element is optional (default)
        :param position: int, this is a positional element, at this position (0 = first position)
        :return: List of allowed elements
        """
        if type(ename) is List:
            for e in ename:
                cls.allow_as_element(
                    e, guaranteed=guaranteed, position=position, also=also
                )
            return
        elif type(ename) is str:
            if guaranteed:
                if ename not in cls._guaranteed:
                    cls._guaranteed.append(ename)
            if position is not None:
                if ename in cls._position:
                    cls._position.remove(ename)
                cls._position.insert(position, ename)
            if ename not in cls.elements_allowed():
                cls._also.append(ename)
        else:
            raise TypeError("first argument must be string or list of strings")
        return

    @classmethod
    def allow_as_part(cls, g: Union[str, type]):
        """Allow g to be enclosed as part of this group.

        Arguments:
            g: the name of a group, or a subclass of KGroup.
               A string or class.
        """
        if g not in cls._part:
            cls._part.append(g)

    @classmethod
    def inc_sequence(cls):
        KGroup._global_sequence += 1
        return KGroup._global_sequence

    @classmethod
    def inc_line(cls):
        KGroup._global_line += 1
        return KGroup._global_line

    def __init__(self, *args, **kwargs):
        """Creates a new instance of group with setting the value of elements.

        Args:
            *args: values for positional arguments
            **kwargs: values for optional elements.

        Example:
            ``ks = KSource("s1", type="test", loc="auc", ref="alumni", obs="Nested")``

        Use element_check=False to turn off checking of element names

        """
        self._containsd: dict = {}
        self.level = 1
        self.line = 1
        self.order = 1
        self._element_check = True
        self._elementsd = {}
        self._inside = None

        if len(args) > len(self._position):
            raise ValueError("Too many positional elements")
        n = 0
        # set the positional arguments according to "_position"
        for arg in args:
            e = self._position[n]
            # setattr(self, e, KElement(e, arg))
            self[e] = arg  # this will go through __setitem__
            n = n + 1
        # keyword arguments must be in one the element lists
        for k, v in kwargs.items():
            if k == "element_check":
                self._element_check = v
            else:
                if self._element_check and not self.is_allowed_as_element(k):
                    raise ValueError(f"Element not allowed: {k}")
                self[k] = v
        # test if the compulsory (guaranteed) elements are present
        for g in self._guaranteed:
            if getattr(self, g, None) is None:
                raise TypeError(
                    f"{self.kname}: element {g} in _guaranteed "
                    f"is missing or with None value"
                )

    def include(self, group: Type["KGroup"]):
        """Include a group. `group`, or its class, must in _part list or
            extend a class in the part list.

        Returns self so it is possible to chain: g.include(g2).include(g3)"""

        allowed = self.is_allowed_as_part(group)
        if allowed is None:
            raise ValueError(f"Group {self.kname} cannot contain {group.kname}")

        group.level = self.level + 1
        group.line = KGroup.inc_line()
        group.order = KGroup.inc_sequence()

        # Hook to before input processing in the group being included
        if hasattr(group, "before_include") and callable(group.before_include):
            if not group.before_include(self):
                raise TypeError(
                    f"{group} includding aborted by "
                    f"group.before_include(self)"
                    f" returning False"
                )

        # new style, dictionary based
        k = self._containsd.keys()
        if allowed in k:
            self._containsd[allowed].append(group)
        else:
            self._containsd[allowed] = [group]
        group.inside = self
        # Hook to after input processing in the group being included
        if hasattr(group, "after_include") and callable(group.after_include):
            group.after_include(self)
        return self

    def before_include(self, container_group):
        """Method called before this group is included into another
        through KGroup.include(KGroup)."""

        if self.id is None:
            if container_group.id is None:
                raise ValueError(
                    "A group with no id cannot be included in another "
                    "group also without id"
                )
            gid = f"{container_group.id}-{self.order:02d}-{self.kname[0:3]}"
            self["id"] = gid
            return True

        return True

    def after_include(self, group):
        """Method called after a new group is included in this one
        through KGroup.include(KGroup)."""
        pass

    def is_allowed_as_part(self, group):
        """Test if a group can be included in the current one.

        For a group to be allowed for inclusion one of 3 conditions necessary:
            1. the kname of the group is in self._pars
            2. the type of the group is in self._pars
            3. the type of the group inherits from a type in self._pars

        Return key under which the group is allowed (kname, type or super type)
        Return None if not allowed
        """
        if not self.is_kgroup(group):
            raise TypeError("Argument must be subclass of KGroup")
        if group.kname not in self._part:
            allowed_classes = [c for c in self._part if type(c) is not str]
            super_classes = type(group).mro()
            r = list(set(super_classes).intersection(set(allowed_classes)))
            if len(r) == 0:
                allowed = None
            else:
                allowed = r[0]  # kname?
        else:
            allowed = group.kname

        return allowed

    def is_allowed_as_element(self, element_name):
        """
        Test if this element is allowed in this group.

        For an element to be allowed one the following must be true:

            1. part of KGroup._builtin
            2. part of position list
            3. part of guarenteed list
            4. part of also list

        Note that this function is unaffected by self._element_check

        :param element_name: name of element to check

        :return: True if element allowed False otherwise
        """
        all_elements = (
            self._builtin_elements + self._position + self._guaranteed + self._also
        )
        return element_name in all_elements

    def includes(self, group: Type[Union[str, Type["KGroup"]]] = None) -> list:
        """Returns included groups.

        Groups are returned by the order in _pars.

        :todo:  this would better be a generator, yield instead of extend.

        :param str group: filter by group name
        """
        if group is not None:
            if group in self._containsd.keys():
                return self._containsd[group]
            else:
                if type(group) is str:
                    gname = group
                elif KGroup.is_kgroup(group):
                    gname = group.kname

                inc_by_part_order = []
                classes_in_contains = [
                    c for c in self._containsd.keys() if hasattr(c, "_name")
                ]
                for class_in_contains in classes_in_contains:
                    if class_in_contains._name == gname:
                        inc_by_part_order.extend(self._containsd[class_in_contains])
                return inc_by_part_order
        else:  # no specific subgroup, we return by pars order
            inc_by_part_order = []
            for p in self._part:
                if p in self._containsd.keys():
                    inc_by_part_order.extend(self._containsd[p])

            return inc_by_part_order

    def attr(
        self,
        the_type: Union[str, KElement, Tuple[str, str, str]],
        value: Union[str, KElement, Tuple[str, str, str]],
        date: Union[str, KElement, Tuple[str, str, str]],
        obs=None,
    ):
        """Utility function to include a KAttribute in this KGroup

        The call::

            KGroup.attr('age','25','2021-08-08',obs='in May')

        is short hand for::

            KGroup.include(KAttr('age','25','2021-08-08',obs='in May'))

        Params google style

        :param str or tuple the_type: core or (core,org,comment)
        :param str or tuple value: core or (core,org,comment)
        :param str date: date as string in Kleio format, or (date,org,comment)
        :param str obs: options observation field

        """
        ka = self._attr_class
        self.include(ka(the_type, value, date=date, obs=obs))
        return self

    def rel(
        self,
        the_type: Union[str, tuple],
        value: Union[str, tuple],
        destname: Union[str, tuple],
        destination: Union[str, tuple],
        date: Union[str, tuple] = None,
        obs: str = None,
    ):
        """include a relation in this KGroup"""
        kr = self._rel_class
        self.include(kr(the_type, value, destname, destination, date=date, obs=obs))

    def to_kleio(self, indent="") -> str:
        """Return a kleio representation of the group."""
        return self.__str__(indent=indent, recurse=True)

    def to_dict(
        self,
        allow_none: bool = False,
        include_str: bool = False,
        include_kleio: bool = False,
        redundant_keys: bool = True,
        include_builtin: bool = True,
    ):
        """Return group information as a dict.

        Params:
            allow_none bool = Include null values (default False)

            :param allow_none: Include null values (default False)
            :param include_str: include a string represention of each element
                                with keys [elementname]_str, the value will include
                                # and % if necessary
            :param include_kleio: include a kleio representation of each element
                                  with keys [elementname]_kleio; this is similar
                                  to *include_str* but the value also contains the
                                  element name
            :param redundant_keys: include redundant keys (see bellow)
            :param include_builtin: include also elements like line,level,... etc...
            :return: a dictionary with the structure bellow

            *Keys of returned dict*

            :group[element]:
                core value of element; also group[element_core]
            :group[element_comment]:
                comment aspect of element
            :group[element_original]:
                original aspect of element
            :group[element_str]:
                 if *include_str=True*, a string representation of elementm(with # and % if necessary);
            :group[element_kleio]:
                if *include_kleio=True*, a kleio representation in the form ``element=string``
            :group[includes]:
                list of enclosed groups
            :group[includes][subgroup]:
                list of enclosed groups of type subgroup


        If *redundant_keys=True* enclosed subgroups can also be accessed
        in the plural form, if no name conflict with existing elements::

                group['persons'] == group['includes']['persons']
                group['person']['id1'] ==
                        [p for p in group['includes']['persons']
                        if p.id='id1'][0]



        :todo: add first(), last() where includes is allowed::
                    group['first']['person'] = group['includes']['person'][0]

        """
        kd = dict()
        kd["kleio_group"] = self._name
        elements_to_include = self.elements_allowed()
        if include_builtin:
            els = elements_to_include.union(set(self._builtin_elements)) - set(
                ["inside"]
            )
        for e in els:
            v: KElement = getattr(self, e, None)
            if v is not None:
                if issubclass(type(v), KElement):
                    core, comment, original = v.to_tuple()
                    kd[e] = core
                    kd[e + "_core"] = core
                    kd[e + "_comment"] = comment
                    kd[e + "_original"] = original
                    if include_str:
                        kd[e + "_str"] = str(v)
                    if include_kleio:
                        kd[e + "_kleio"] = v.to_kleio()
                else:
                    kd[e] = v
        if not allow_none:
            kd = dict([(key, value) for key, value in kd.items() if value is not None])
        # we now includes subgroups
        ki = dict()
        # we now collect subgroups by name
        included = list(self.includes())
        for i in included:
            n = i.kname
            if n not in ki.keys():
                ki[n] = [
                    i.to_dict(
                        include_str=include_str,
                        include_kleio=include_kleio,
                        include_builtin=include_builtin,
                        redundant_keys=redundant_keys,
                    )
                ]
            else:
                ki[n].append(
                    i.to_dict(
                        include_str=include_str,
                        include_kleio=include_kleio,
                        include_builtin=include_builtin,
                        redundant_keys=redundant_keys,
                    )
                )
        if len(ki) > 0:
            kd["includes"] = ki
            # if there are no name conflicts and plural form
            # so g['includes']['act'] can be accessed as
            #    g['acts']
            if redundant_keys:
                for subgroup in ki.keys():
                    if subgroup + "s" not in self.elements_allowed():
                        kd[subgroup + "s"] = ki[subgroup]
                        # we include subgroup indexed by id
                        # so we can have source['act']['ac010]['person']['p01']
                        for group in ki[subgroup]:
                            gid = group.get("id", None)
                            if (
                                gid is not None
                                and subgroup not in self.elements_allowed()  # noqa
                            ):
                                if subgroup not in kd.keys():
                                    kd[subgroup] = dict()
                                kd[subgroup][gid] = group
        return kd

    def to_json(self):
        return json.dumps(
            self.to_dict(
                include_str=False,
                include_kleio=False,
                redundant_keys=False,
                include_builtin=True,
            ),
            indent=4,
            allow_nan=False,
        )

    @property
    def get(self):
        return self.to_dict()

    def to_dots(self):
        return Box(self.to_dict())

    @property
    def dots(self):
        """
        Allows easy referencing of the dictionary representation of the group.

        It is very usefull in list comprehensiona, e.g.:

          >> 'Diamantina' in [ls.value for ls in person.dots.lss
                                if ls.type == 'nome-geografico']

          >> [ls.value for ls in person.dots.lss
                        if ls.type == 'nome-geografico']

        :return:
        """
        return self.to_dots()

    def __str__(self, indent="", recurse=False):
        sname = getattr(self, "_name", self.__class__.__name__)
        s = str(sname) + "$"
        first = True
        out = []
        for e in self._position:
            v: KElement = getattr(self, e, None)
            if v is not None:
                if not first:
                    s = s + "/" + v.to_kleio(name=False)
                else:
                    s = s + v.to_kleio(name=False)
                    first = False
                out.append(e)
        more = sorted(
            list(
                set(self._guaranteed)
                .union(set(self._also))
                .union(self._position)
                .difference(out)
            )
        )
        # print(more)
        if "obs" in more:  # we like obs elements at the end
            more.remove("obs")
            more.append("obs")
        for e in more:
            m: Union[KElement, str] = getattr(self, e, None)
            if m is not None and (
                type(m) is str
                and m > ""  # noqa
                or (issubclass(type(m), KElement) and m.to_kleio() > "")  # noqa
            ):
                # m contains data, lets output
                if issubclass(type(m), KElement):
                    v = m.to_kleio()
                else:
                    v = KElement(e, m).to_kleio()
                if not first:
                    s = s + f"/{v}"
                else:
                    s = s + f"{v}"
                    first = False

        if recurse:
            for g in self.includes():
                s = s + nl + g.__str__(indent + " ", recurse=recurse)
        return textwrap.indent(s, indent)

    def __getitem__(self, arg):
        if self._element_check and arg not in self.elements_allowed():
            raise ValueError("Element does not exist in group")
        return getattr(self, arg)

    def __setitem__(self, arg, value):
        if self._element_check and not self.is_allowed_as_element(arg):
            raise ValueError(f"Element not allowed: {arg}")
        el = self.pack_as_kelement(arg, value)
        setattr(self, arg, el)
        self._elementsd[arg] = el

    def pack_as_kelement(self, arg, value, element_class=None):
        """Packs value as a KElement with name arg

        :param arg: name of element
        :param value: value of element; can be a KElement, a tuple
                      (core, cooment, original) or a value
        :param element_class: class of element as string; if none arg is used to find class

        TODO: the question is what is the class of the KElement to be stored
        if value is a KElement then it already has a class
        """
        # TODO: this is wrong. if value is a KElement it should be stored as such

        if element_class is None:
            kelement = KElement.get_class_for(arg)
        else:
            kelement = KElement.get_class_for(element_class)
        if kelement is None and isinstance(value, KElement):
            if isinstance(value.element_class, KElement):
                kelement = value.element_class
            else:
                kelement = KElement.get_class_for(value.element_class)
        if kelement is None:  # if there is no KElement class we create it
            kelement = KElement.extend(arg)
            warnings.warn(
                f"Created a KElement class for {arg}. "
                f"Better to create explicit or provide "
                f" synonyms= in group creation.",
                stacklevel=2,
            )
            # we get KElement class that matches the name
            # this is how we handle localized name of elements that
            # have a builtin meaning or are referred to by standard names
            # in PomSomMapping
        if isinstance(value, KElement):
            el = kelement(
                core=value.core, comment=value.comment, original=value.original
            )
        else:
            comment = None
            original = None
            if type(value) is tuple and len(value) == 3:
                core, comment, original = value
            else:
                core = value
            el = kelement(arg, core=core, comment=comment, original=original)
        return el

    def unpack_from_kelement(self, value):
        """
        if value is a KElement return core if not return value as is.

        Useful to obtain the core value in elements that normally have
        no comment or original.

        :param value:
        :return:
        """
        if not isinstance(value, KElement):  # we did not get a KElement
            return value
        else:  # we got a KElement object
            return value.core

    def get_core(self, element, default=None):
        """get_core(element_name [, default])
        Returns the core value of an element
        """
        try:
            core = self.unpack_from_kelement(self[element])
        except ValueError:
            core = None
        except AttributeError:
            core = None
        if core is None:
            return default
        else:
            return core

    def get_id(self):
        """Return the id of the group"""
        return self.unpack_from_kelement(self.id)

    def get_element_for_column(self, colspec, default=None):
        """
        Return the value of an element that matches a specific column in the
        database.

        An element matches a column if its name is equal to the column name
        or if it is an instance of KElement subclass with the same name.

        :param colspec: name of column, or name of POMClassAttributes
        :param default: default value if not element found
        :return: KElement, or whatever in default.

        """
        el: KElement
        for (
            name,
            el,
        ) in (
            self._elementsd.items()
        ):  # same name as column or in ancestors # noqa: E501
            if name == colspec:
                return el
        # Handles synonyms created by subclassing core KElements
        for el in self._elementsd.values():  # if name in inherited names
            if colspec in el.inherited_names():
                return el
        # handles multiple subclassing of core KElements
        for (
            el
        ) in (
            self._elementsd.values()
        ):  # check if there are alternative classes # noqa: E501
            # all classes for colspec
            targets = KElement.get_classes_for(colspec)
            # other classes for el
            alternatives = KElement.get_classes_for(el.name)
            # now check if there is a common path
            #   collect all the ancestors of all the classes for this el
            alt_ancestors = [(sc, sc.inherits_from()) for sc in alternatives]
            #   check if any of colspec classes are there
            for _alternative, ancestors in alt_ancestors:
                common = set(ancestors).intersection(set(targets))
                if len(common) > 0:
                    return el
        return default



import os

import pytest  # pylint: disable=import-error
from sqlalchemy import inspect  # noqa

from tests import skip_on_travis, conn_string
from timelink.kleio.groups import KElement, KGroup, KSource, KAct, KPerson
from timelink.api.models import base  # noqa
from timelink.api.models.base_class import Base
from timelink.api.database import TimelinkDatabase
from timelink.api.models.entity import Entity  # noqa
from timelink.api.models.pom_som_mapper import PomSomMapper
from timelink.api.models.source import Source

# pytestmark = skip_on_travis


@pytest.fixture(scope="module")
def get_db():
    """Returns a database connection"""
    database = TimelinkDatabase(db_url=conn_string)
    try:
        yield database
    finally:
        with database.session() as db:
            database.drop_db(db)


@pytest.fixture
def kgroup_person_attr_rel() -> KSource:
    """Returns a group wtih attr and rel"""
    ks = KSource(
        "s2",
        type="test",
        loc="auc",
        ref="attr-rel",
        obs="""
        This is a multiline obs with special characteres like /,= and $
        """,
    )
    ka1 = KAct(
        "a2-1",
        "test-act",
        date="2021-07-16",
        day=16,
        month=7,
        year=2021,
        loc="macau",
        ref="p.1",
        obs="Test Act",
    )
    ks.include(ka1)
    p1 = KPerson("Joaquim", "m", "p01-2", obs="Living in Macau/China")
    p1.attr("residencia", "Macau", date="2021-12-11")
    mobs = """
    Moving from Portugal
    to Macau in 2022
    """
    p2 = KPerson("Margarida", "f", "p02-2", obs=mobs)
    p2.attr("residencia", "Trouxemil", date="2020-10-18")
    p1.rel(
        "parentesco", "marido", p2.name, p2.id, date="2006-01-4", obs="Ilha Terceira"
    )
    ka1.include(p2)
    ka1.include(p1)
    return ks


@pytest.fixture
def kgroup_nested() -> KSource:
    """Returns a nested Kleio structure

    This is a nested Kleio Group Source/Act/Person/Attribute
    """
    ks = KSource("s1", type="test", loc="auc", ref="alumni", obs="Nested")
    ka1 = KAct(
        "a1",
        "test-act",
        date="2021-07-16",
        day=16,
        month=7,
        year=2021,
        loc="auc",
        ref="p.1",
        obs="Test Act",
    )
    ks.include(ka1)
    ka2 = KAct(
        "a2",
        "test-act",
        date="2021-07-17",
        day=17,
        month=7,
        year=2021,
        loc="auc",
        ref="p.2",
        obs="Test Act",
    )

    ks.include(ka2)
    p1 = KPerson("Joaquim", "m", "p01")
    p1.attr("residencia", "Macau", date="2021-12-11")
    p2 = KPerson("Margarida", "f", "p02")
    p2.attr("residencia", "Trouxemil", date="2020-10-18")
    p3 = KPerson("Pedro", "m", "p03")
    p3.attr("residencia", "Coimbra", date="2021-10-21")

    ka1.include(p1)
    ka1.include(p2)
    ka1.include(p3)
    p4 = KPerson("Maria", "f", "p04")
    p5 = KPerson("Manuel", "m", "p05")
    p6 = KPerson("João", "m", "p06")
    ka2.include(p4)
    ka2.include(p5)
    ka2.include(p6)
    return ks


@skip_on_travis
def test_succeed_if_not_in_travis():
    assert os.getenv("TRAVIS") != "true"


def test_create_db(get_db):
    # Database is created and initialized in the fixture
    metadata = Base.metadata
    tables = list(metadata.tables.keys())
    assert len(tables) > 0, "tables where not created"


def test_create_nattribute(get_db, kgroup_nested):
    # Database is created and initialized in the fixture
    db: TimelinkDatabase = get_db
    ks = kgroup_nested
    with get_db.session() as session:
        PomSomMapper.store_KGroup(ks, session)
        session.commit()

    # views = inspect(db.get_engine()).get_view_names()
    # l1 = len(views)
    nattr = db.get_nattribute_view()
    # l2 = len(inspect(get_db.get_engine()).get_view_names())
    assert nattr is not None


def test_entity_contains(get_db):
    """Test entity contains relationship

    In this test four entities are created
    with one containing two others and one of the contained
    entities containing the fourth entity.

    Note that adding the top level entity
    to the session will add all entities to the session
    and deleting the top level entity will delete all entities.

    """
    ent1: Entity = Entity(id="e001", groupname="entity")
    ent1.pom_class = "entity"
    ent1.line = 1
    ent1.level = 1
    ent1.sequence = 1

    ent2: Entity = Entity(id="e002")
    ent2.groupname = "entity"
    ent2.pom_class = "entity"
    ent2.line = 2
    ent2.level = 2
    ent2.sequence = 2
    ent1.contains.append(ent2)

    ent3: Entity = Entity(id="e003")
    ent3.groupname = "entity"
    ent3.pom_class = "entity"
    ent3.line = 2
    ent3.level = 2
    ent3.sequence = 2
    ent1.contains.append(ent3)

    ent4: Entity = Entity(id="e004")
    ent4.groupname = "entity"
    ent4.pom_class = "entity"
    ent4.line = 4
    ent4.level = 4
    ent4.sequence = 4
    ent2.contains.append(ent4)

    with get_db.session() as session:
        session.add(ent1)  # this adds the four entities to the database
        session.commit()
        ent4 = ent2.contained_by  # pylint: disable=no-member
        assert ent4 is ent1
        ent5 = ent1.contains[1]
        assert ent5 is ent3
        session.delete(ent1)  # should delete ent2, ent3 and ent4
        session.commit()
        deleted = session.get(Entity, "e004")
        assert deleted is None, "Should have deleted orphan"


def test_ensure_mapping(get_db):
    """Test that all classes registered in the classes table are mapped to ORM

    During database initialization, the classes and class_attributes tables
    is populated with definitions of base classes. Also in the timelink.api.models
    package, the base classes are defined as ORM classes. This test checks that
    all classes in the classes table are mapped to ORM classes.

    """
    with get_db.session() as session:
        pom_classes = PomSomMapper.get_pom_classes(session=session)
        pom_ids = PomSomMapper.get_pom_class_ids(session=session)
        pom_tables = [pom_class.table_name for pom_class in pom_classes]
        pom_class: PomSomMapper
        for pom_class in pom_classes:
            pom_class.ensure_mapping(session=session)
            # print(repr(pom_class))
            # print(pom_class)
        orm_mapped_classes = Entity.get_som_mapper_to_orm_as_dict()
        non_mapped = set(pom_ids) - set(orm_mapped_classes.keys())
        assert len(non_mapped) == 0, "Not all classes are mapped to ORM"
        orm_mapped_tables = Entity.get_tables_to_orm_as_dict()
        tables_not_mapped = set(pom_tables) - set(orm_mapped_tables)
        assert len(tables_not_mapped) == 0, "Not all class tables are mapped in ORM"
        db_tables = get_db.table_names()
        tables_not_in_db = set(orm_mapped_tables) - set(db_tables)
        assert (
            len(tables_not_in_db) == 0
        ), "Not all mapped tables were created in the db"
        session.close()


def test_insert_nested_groups(get_db, kgroup_nested):
    """Test inserting a nested Kleio group"""
    ks = kgroup_nested
    source_id = ks.get_id()
    with get_db.session() as session:
        PomSomMapper.store_KGroup(ks, session)
        session.commit()
        source_from_db: Source = Entity.get_entity(source_id, session)
        assert source_from_db.id == ks.get_id()
        assert repr(source_from_db)
        assert str(source_from_db)
        acts = source_from_db.contains
        act = acts[0]
        ract = repr(act)
        sact = str(act)
        assert ract
        assert sact
        person = act.contains[0]
        assert repr(person)
        assert str(person)


def test_insert_entities_attr_rel(get_db, kgroup_person_attr_rel):
    """Test inserting a group with attributes and relationships"""
    ks = kgroup_person_attr_rel
    source_id = ks.get_id()
    with get_db.session() as session:
        PomSomMapper.store_KGroup(ks, session)
        session.commit()
        source_from_db = Entity.get_entity(source_id, session)
        assert source_from_db.id == ks.get_id()


def test_quote_and_long_test(kgroup_person_attr_rel):
    ks = kgroup_person_attr_rel
    kleio = ks.to_kleio()
    assert '"""' in kleio, "Bad handling of long text"


def test_insert_delete_previous_source(get_db, kgroup_nested):
    ks = kgroup_nested
    source_id = ks.get_id()
    with get_db.session() as session:
        print(
            f"""

        =================
        Inserting a nested group

        {ks.to_kleio()}
        """
        )
        PomSomMapper.store_KGroup(ks, session)
        session.commit()
        ks2 = KSource(
            "s1",
            type="test",
            loc="auc",
            ref="alumni",
            obs="Empty should delete previous one",
        )

        print(
            """
        ================
        Reinserting group with same id
        ===============
        """
        )
        PomSomMapper.store_KGroup(ks2, session)
        source_from_db: Entity = Entity.get_entity(source_id, session)
        contains = source_from_db.contains
        assert len(contains) == 0, "import did not delete contained entities"


def test_store_KGroup_1(get_db):
    """Test storing a group with non core group"""
    kfonte: KGroup = KSource.extend(
        "fonte",
        position=["id", "data"],
        also=["tipo", "ano", "obs", "substitui"],
        synonyms=[
            ("tipo", "type"),
            ("ano", "year"),
            ("data", "date"),
            ("substitui", "replace"),
        ],
    )
    afonte = kfonte(
        "f001",
        data=KElement("data", "2021-12-02", element_class="date"),
        tipo=KElement("tipo", "teste", element_class="type"),
        obs="First group stored through ORM with non core group",
    )
    with get_db.session() as session:
        session.commit()
        # this converts a group to an database entity
        afonte_entity = PomSomMapper.kgroup_to_entity(afonte, session)
        assert afonte_entity is not None
        PomSomMapper.store_KGroup(afonte, session)
        session.commit()
        source_orm = Entity.get_orm_for_pom_class(afonte.pom_class_id)
        same_fonte = session.get(source_orm, afonte.id.core)
        assert str(afonte.obs) == str(same_fonte.obs)  # noqa
        assert str(afonte.data) == str(same_fonte.the_date)  # noqa
        session.close()



#!/usr/bin/env python

"""Tests for `timelink-py` package."""

import pytest
from typer.testing import CliRunner

from timelink.cli import app
from tests import mhk_absent


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "Timelink and MHK manager" in result.output


@mhk_absent
def test_command_mhk_version():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(app, ["mhk", "version"])
    print(f"Test result {result.exit_code}")
    # assert result.exit_code == 0
    assert "Portainer URL" in result.output or "not found" in result.output


@mhk_absent
def test_command_mhk_status():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(app, ["mhk", "status"])
    assert result.exit_code == 0
    assert "Stopped" in result.output



""" These mappings are needed to boostrap a new database.

They are used by TimelinkDatabase to initialize a new database.

Mappings as these can be generated from existing Timelink/MHK databases with:

.. code-block:: python

   pom_classes = session.query(PomSomMapper)
                        .where(Entity.pom_clas == 'class').all()
   for pom_class in pom_classes:
        print(f"'{pom_class.id}': [")
        print(repr(pom_class),',')
        for cattr in pom_class.class_attributes:
            print(repr(cattr),',')
        print('],')

Note that the order of the mappings is important. If a mapping
extends another mapping with 'super_class', the super class mapping must be defined
before the subclass mapping.

"""

from timelink.api.models.pom_som_mapper import PomSomMapper, PomClassAttributes

pom_som_base_mappings = {
    "entity": [
        PomSomMapper(
            id="entity", table_name="entities", group_name="na", super_class="root"
        ),
        PomClassAttributes(
            the_class="entity",
            name="class",
            colname="class",
            colclass="class",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="entity",
            name="groupname",
            colname="groupname",
            colclass="groupname",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="entity",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="entity",
            name="inside",
            colname="inside",
            colclass="inside",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="entity",
            name="level",
            colname="the_level",
            colclass="level",
            coltype="numeric",
            colsize="3",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="entity",
            name="line",
            colname="the_line",
            colclass="line",
            coltype="numeric",
            colsize="6",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="entity",
            name="order",
            colname="the_order",
            colclass="order",
            coltype="numeric",
            colsize="6",
            colprecision="0",
            pkey="0",
        ),
    ],
    "act": [
        PomSomMapper(
            id="act",
            table_name="acts",
            group_name="historical-act",
            super_class="entity",
        ),
        PomClassAttributes(
            the_class="act",
            name="date",
            colname="the_date",
            colclass="date",
            coltype="varchar",
            colsize="24",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="act",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="act",
            name="loc",
            colname="loc",
            colclass="loc",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="act",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="32768",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="act",
            name="ref",
            colname="ref",
            colclass="ref",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="act",
            name="type",
            colname="the_type",
            colclass="type",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
    ],
    "attribute": [
        PomSomMapper(
            id="attribute",
            table_name="attributes",
            group_name="attribute",
            super_class="entity",
        ),
        PomClassAttributes(
            the_class="attribute",
            name="date",
            colname="the_date",
            colclass="date",
            coltype="varchar",
            colsize="24",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="attribute",
            name="entity",
            colname="entity",
            colclass="entity",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="attribute",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="attribute",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="32768",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="attribute",
            name="type",
            colname="the_type",
            colclass="type",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="attribute",
            name="value",
            colname="the_value",
            colclass="value",
            coltype="varchar",
            colsize="254",
            colprecision="0",
            pkey="0",
        ),
    ],
    "geoentity": [
        PomSomMapper(
            id="geoentity",
            table_name="geoentities",
            group_name="geoentity",
            super_class="entity",
        ),
        PomClassAttributes(
            the_class="geoentity",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="geoentity",
            name="name",
            colname="name",
            colclass="name",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="geoentity",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="32768",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="geoentity",
            name="type",
            colname="the_type",
            colclass="type",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
    ],
    "object": [
        PomSomMapper(
            id="object", table_name="objects", group_name="object", super_class="entity"
        ),
        PomClassAttributes(
            the_class="object",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="object",
            name="name",
            colname="name",
            colclass="name",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="object",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="32768",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="object",
            name="type",
            colname="the_type",
            colclass="type",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
    ],
    "good": [
        PomSomMapper(
            id="good", table_name="goods", group_name="bem", super_class="object"
        ),
        PomClassAttributes(
            the_class="good",
            name="description",
            colname="description",
            colclass="description",
            coltype="varchar",
            colsize="1024",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="good",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="good",
            name="loc",
            colname="loc",
            colclass="loc",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
    ],
    "person": [
        PomSomMapper(
            id="person", table_name="persons", group_name="person", super_class="entity"
        ),
        PomClassAttributes(
            the_class="person",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="person",
            name="name",
            colname="name",
            colclass="name",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="person",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="32768",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="person",
            name="sex",
            colname="sex",
            colclass="sex",
            coltype="char",
            colsize="1",
            colprecision="0",
            pkey="0",
        ),
    ],
    "relation": [
        PomSomMapper(
            id="relation",
            table_name="relations",
            group_name="relation",
            super_class="entity",
        ),
        PomClassAttributes(
            the_class="relation",
            name="date",
            colname="the_date",
            colclass="date",
            coltype="varchar",
            colsize="24",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="relation",
            name="destination",
            colname="destination",
            colclass="destination",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="relation",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="relation",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="32768",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="relation",
            name="origin",
            colname="origin",
            colclass="origin",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="relation",
            name="type",
            colname="the_type",
            colclass="type",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="relation",
            name="value",
            colname="the_value",
            colclass="value",
            coltype="varchar",
            colsize="254",
            colprecision="0",
            pkey="0",
        ),
    ],
    "rentity": [
        PomSomMapper(
            id="rentity",
            table_name="rentities",
            group_name="rentity",
            super_class="entity",
        ),
        PomClassAttributes(
            the_class="rentity",
            name="description",
            colname="description",
            colclass="name",
            coltype="varchar",
            colsize="128",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="rentity",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="rentity",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="32768",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="rentity",
            name="status",
            colname="status",
            colclass="status",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="rentity",
            name="the_class",
            colname="the_class",
            colclass="the_class",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="rentity",
            name="user",
            colname="user",
            colclass="user",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
    ],
    "rgeoentity": [
        PomSomMapper(
            id="rgeoentity",
            table_name="rgeoentities",
            group_name="rgeoentity",
            super_class="rentity",
        ),
        PomClassAttributes(
            the_class="rgeoentity",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="rgeoentity",
            name="sname",
            colname="sname",
            colclass="sname",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
    ],
    "robject": [
        PomSomMapper(
            id="robject",
            table_name="robjects",
            group_name="robject",
            super_class="rentity",
        ),
        PomClassAttributes(
            the_class="robject",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="robject",
            name="sname",
            colname="sname",
            colclass="sname",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="robject",
            name="type",
            colname="the_type",
            colclass="type",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
    ],
    "rperson": [
        PomSomMapper(
            id="rperson",
            table_name="rpersons",
            group_name="rperson",
            super_class="rentity",
        ),
        PomClassAttributes(
            the_class="rperson",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="rperson",
            name="sex",
            colname="sex",
            colclass="sex",
            coltype="char",
            colsize="1",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="rperson",
            name="sname",
            colname="sname",
            colclass="sname",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
    ],
    "source": [
        PomSomMapper(
            id="source",
            table_name="sources",
            group_name="historical-source",
            super_class="entity",
        ),
        PomClassAttributes(
            the_class="source",
            name="date",
            colname="the_date",
            colclass="date",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="source",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="source",
            name="kleiofile",
            colname="kleiofile",
            colclass="kleiofile",
            coltype="varchar",
            colsize="512",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="source",
            name="loc",
            colname="loc",
            colclass="loc",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="source",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="32768",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="source",
            name="ref",
            colname="ref",
            colclass="ref",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="source",
            name="type",
            colname="the_type",
            colclass="type",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="source",
            name="value",
            colname="the_value",
            colclass="value",
            coltype="varchar",
            colsize="254",
            colprecision="0",
            pkey="0",
        ),
    ],
}

more_mappings = {
    "acusacoes": [
        PomSomMapper(
            id="acusacoes",
            table_name="acusacoes",
            group_name="acusa",
            super_class="object",
        ),
        PomClassAttributes(
            the_class="acusacoes",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="acusacoes",
            name="idcaso",
            colname="idcaso",
            colclass="idcaso",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="acusacoes",
            name="literal",
            colname="literal",
            colclass="literal",
            coltype="varchar",
            colsize="16000",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="acusacoes",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="16000",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="acusacoes",
            name="origem",
            colname="origem",
            colclass="origem",
            coltype="varchar",
            colsize="16000",
            colprecision="0",
            pkey="0",
        ),
    ],
    "caso": [
        PomSomMapper(
            id="caso", table_name="casos", group_name="caso", super_class="object"
        ),
        PomClassAttributes(
            the_class="caso",
            name="id",
            colname="id",
            colclass="id",
            coltype="varchar",
            colsize="64",
            colprecision="0",
            pkey="1",
        ),
        PomClassAttributes(
            the_class="caso",
            name="obs",
            colname="obs",
            colclass="obs",
            coltype="varchar",
            colsize="16654",
            colprecision="0",
            pkey="0",
        ),
        PomClassAttributes(
            the_class="caso",
            name="type",
            colname="the_type",
            colclass="the_type",
            coltype="varchar",
            colsize="32",
            colprecision="0",
            pkey="0",
        ),
    ],
}
